import cerberus
import datetime
import json
from munch import Munch
from openamundsen import constants, util
from openamundsen.errors import ConfigurationError
import pandas as pd
from pathlib import Path
import re

try:
    import openamundsen_snowmanagement
    SNOW_MANAGEMENT_AVAILABLE = True
    SNOW_MANAGEMENT_DATA_DIR = Path(openamundsen_snowmanagement.__file__).parent / 'data'
except ImportError:
    SNOW_MANAGEMENT_AVAILABLE = False

DATA_DIR = Path(__file__).parent / 'data'


class ConfigurationValidator(cerberus.Validator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _normalize_coerce_float(self, value):
        return float(value)

    def _normalize_coerce_datetime(self, date):
        return pd.to_datetime(date)

    def _normalize_coerce_path(self, path):
        return Path(path)

    def _normalize_coerce_timestep(self, timestep):
        # As of pandas 2.2.0, some frequency aliases have been deprecated (see
        # https://pandas.pydata.org/docs/dev/whatsnew/v2.2.0.html). Since especially the "H" alias
        # (which is now deprecated in favor of "h") has been frequently used in earlier versions of
        # openAMUNDSEN, we replace it here manually with its lowercase version in order to avoid
        # deprecation warnings.
        if timestep.endswith('H'):
            timestep = timestep[:-1] + 'h'
        return timestep

    def _normalize_coerce_snowmodel(self, model):
        # Snow model name 'layers' is deprecated -> change to 'multilayer'
        if model == 'layers':
            return 'multilayer'

        return model

    def _normalize_coerce_meteo_interpolation(self, d):
        # 'wind_speed' is deprecated -> change to 'wind'
        if 'wind_speed' in d:
            d['wind'] = d.pop('wind_speed')

        return d

    def _normalize_coerce_cloudiness(self, d):
        # 'day_method' is deprecated -> change to 'method'
        if 'day_method' in d:
            d['method'] = d.pop('day_method')

        # 'night_method' is deprecated -> change to 'clear_sky_fraction_night_method'
        if 'night_method' in d:
            d['clear_sky_fraction_night_method'] = d.pop('night_method')

        return d


class ConfigurationEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, pd.Timestamp):
            return str(o)
        else:
            raise TypeError


def parse_end_date(end_date, timestep):
    # If end_date is specified without an hour value, the end hour should be inferred
    # (i.e., set to the latest time step of the end day).
    if type(end_date) is datetime.date:  # do not use isinstance() because we want to catch only datetime.date and not its subclasses (datetime.datetime, pd.Timestamp etc.)
        infer_end_hour = True
    elif isinstance(end_date, str) and re.match(r'^\d\d\d\d-\d\d-\d\d$', end_date.strip()):
        infer_end_hour = True
    else:
        infer_end_hour = False

    end_date = pd.to_datetime(end_date)

    # If no end hour is specified (only the date), set it to the last time step
    # of the respective day (for the start date the hour is automatically set
    # to 0 if not explicitly specified)
    timedelta = util.offset_to_timedelta(timestep)
    if infer_end_hour and timedelta < pd.Timedelta(days=1):
        end_date += pd.Timedelta(hours=24) - timedelta

    return end_date


class Configuration(Munch):
    """
    Container for storing model configuration. This class inherits from `Munch`
    so that attributes are accessible both using dict notation
    (`config['start_date']`) as well as dot notation (`config.end_date`).
    """
    @classmethod
    def from_dict(cls, d):
        return cls.fromDict(d)

    @classmethod
    def from_yaml(cls, s):
        return cls.from_dict(util.load_yaml(s))

    def __repr__(self):
        return util.to_yaml(self.toDict())


def read_config(filename):
    """
    Read a configuration (YAML) file and return the resulting dict as a
    Configuration object.
    """
    return Configuration.fromDict(util.read_yaml_file(filename))


def parse_config(config):
    _init_schemas()
    v = ConfigurationValidator(cerberus.schema_registry.get('openamundsen_config'))
    valid = v.validate(config)

    if not valid:
        raise ConfigurationError('Invalid configuration\n\n' + util.to_yaml(v.errors))

    full_config = Configuration.fromDict(v.document)
    full_config['end_date'] = parse_end_date(full_config['end_date'], full_config['timestep'])

    full_config['land_cover']['classes'] = _merge_land_cover_params(
        full_config['land_cover']['classes']
    )

    validate_config(full_config)

    return full_config


def _init_schemas():
    """
    Read in the validation schemas and add them to the Cerberus registry.
    """
    schema_reg = cerberus.schema_registry
    schemas = schema_reg.all()

    if 'openamundsen_config' not in schemas:
        config_schema = util.read_yaml_file(f'{DATA_DIR}/configschema.yml')

        if SNOW_MANAGEMENT_AVAILABLE:
            sm_schema = util.read_yaml_file(f'{SNOW_MANAGEMENT_DATA_DIR}/configschema.yml')
            config_schema.update(sm_schema)

        schema_reg.add('openamundsen_config', config_schema)

        lcc_schema = util.read_yaml_file(f'{DATA_DIR}/land_cover_class_schema.yml')
        schema_reg.add('openamundsen_land_cover_class', lcc_schema)


def _merge_land_cover_params(class_params):
    """
    Merge the manually set land cover class parameters with the default parameters.
    """
    default_config = read_config(f'{DATA_DIR}/land_cover_class_params.yml')

    dict_schema = {
        'classes': {
            'type': 'dict',
            'valuesrules': {
                'schema': 'openamundsen_land_cover_class',
            },
        },
    }
    v = cerberus.Validator(dict_schema)
    valid = v.validate(default_config)

    if not valid:
        raise ConfigurationError('Invalid land cover configuration\n\n' + util.to_yaml(v.errors))

    default_class_params = Configuration.fromDict(v.document)['classes']
    merged_params = default_class_params.copy()

    for lcc, lcc_params in class_params.items():
        if lcc in merged_params:
            for k, v in lcc_params.items():
                merged_params[lcc][k] = v
        else:
            merged_params[lcc] = lcc_params

    return merged_params


def validate_config(config):
    """
    Perform some additional validations which are too complicated with Cerberus.
    """
    if config.start_date > config.end_date:
        raise ConfigurationError('End date must be after start date')

    # Check if timestep matches start/end dates
    dates = pd.date_range(
        start=config.start_date,
        end=config.end_date,
        freq=config.timestep,
    )
    if dates[-1] != config.end_date:
        raise ConfigurationError('Start/end date is not compatible with timestep')

    # Check if write_freq is compatible with timestep - as long as the time step is <= 1d,
    # write_freq can be an offset like 'ME' or 'YE', but for larger timesteps it is not guaranteed
    # that these dates generated with pd.date_range(start=start_date, end=end_date, freq=write_freq)
    # are actually reached, so in this case write_freq must be a multiple of timestep
    # (e.g. timestep = '5D' and write_freq = '30D')
    timestep_td = util.offset_to_timedelta(config.timestep)
    write_freq = config.output_data.timeseries.write_freq
    if timestep_td > pd.Timedelta(days=1):
        try:
            write_freq_td = pd.Timedelta(write_freq)
            if write_freq_td.total_seconds() % timestep_td.total_seconds() != 0:
                raise ConfigurationError('write_freq must be a multiple of timestep')
        except ValueError:
            raise ConfigurationError('write_freq must be a multiple of timestep')

    if config.input_data.meteo.format == 'netcdf' and config.input_data.meteo.crs is not None:
        print('Warning: Ignoring CRS specification for meteo input data '
              '(CRS not required when using NetCDF format)')

    if config.input_data.meteo.format == 'memory':
        if (
            config.input_data.meteo.bounds != 'grid'
            or len(config.input_data.meteo.exclude) > 0
            or len(config.input_data.meteo.include) > 0
        ):
            print('Warning: "bounds", "exclude" and "include" are currently ignored for '
                  'format "memory"')

    if config.snow.model == 'multilayer' and config.snow.melt.method != 'energy_balance':
        raise ConfigurationError(f'Melt method "{config.snow.melt.method}" not supported for the '
                                 f'snow model "{config.snow.model}"')

    if config.snow.melt.method == 'temperature_index':
        if config.snow.melt.degree_day_factor is None:
            raise ConfigurationError('Missing field: snow.melt.degree_day_factor')
    elif config.snow.melt.method == 'enhanced_temperature_index':
        if config.snow.melt.degree_day_factor is None:
            raise ConfigurationError('Missing field: snow.melt.degree_day_factor')
        if config.snow.melt.albedo_factor is None:
            raise ConfigurationError('Missing field: snow.melt.albedo_factor')

    if abs(config.meteo.precipitation_phase.threshold_temp) < 20:
        print('Warning: precipitation phase threshold temperature seems to be in °C, converting to K')
        config.meteo.precipitation_phase.threshold_temp += constants.T0

    ac = config.snow.albedo
    if ac.method == 'usaco':
        print('Warning: albedo method "usaco" is deprecated, please use "snow_age"')
        ac.method = 'snow_age'
        ac.cold_snow_decay_timescale = 1 / ac.k_neg * constants.HOURS_PER_DAY
        ac.melting_snow_decay_timescale = 1 / ac.k_pos * constants.HOURS_PER_DAY
        ac.decay_timescale_determination_temperature = 'air'
        ac.refresh_snowfall = ac.significant_snowfall
        ac.refresh_method = 'binary'
    elif ac.method == 'fsm':
        print('Warning: albedo method "fsm" is deprecated, please use "snow_age"')
        ac.method = 'snow_age'
        ac.decay_timescale_determination_temperature = 'surface'
        ac.refresh_method = 'continuous'

    if SNOW_MANAGEMENT_AVAILABLE:
        openamundsen_snowmanagement.validate_config(config)



import numpy as np
from openamundsen.fileio import read_raster_file
import pytest

RASTER_TEMPLATE = '''
ncols        6
nrows        5
xllcorner    630800
yllcorner    5180500
cellsize     1000
nodata_value {val_nodata}
{val1} {val1} {val1} {val1} {val1} {val1}
{val1} {val1} {val2} {val2} {val2} {val2}
{val_nodata} {val2} {val2} {val2} {val2} {val2}
{val1} {val1} {val2} {val2} {val2} {val2}
{val1} {val_nodata} {val2} {val2} {val2} {val2}
'''.strip()


def test_fill_value(tmp_path):
    fn = tmp_path / 'test.asc'
    with open(fn, 'w') as f:
        f.write(RASTER_TEMPLATE.format(val1=0, val2=1, val_nodata=-9999))

    data = read_raster_file(fn)
    assert data[2, 0] == -9999

    data = read_raster_file(fn, fill_value=-1)
    assert data[2, 0] == -1

    with pytest.raises(TypeError):
        data = read_raster_file(fn, fill_value=np.nan)

    data = read_raster_file(fn, fill_value=np.nan, dtype=float)
    assert np.isnan(data[2, 0])

    data = read_raster_file(fn, fill_value=False, dtype=bool)
    assert not data[2, 0]


def test_dtype(tmp_path):
    fn = tmp_path / 'test.asc'
    with open(fn, 'w') as f:
        f.write(RASTER_TEMPLATE.format(val1=0, val2=1, val_nodata=-9999))

    data = read_raster_file(fn)
    assert np.issubdtype(data.dtype, np.integer)

    with pytest.raises(TypeError):
        data = read_raster_file(fn, fill_value=np.nan)

    data = read_raster_file(fn, fill_value=np.nan, dtype=float)
    assert np.issubdtype(data.dtype, float)

    data = read_raster_file(fn, fill_value=False, dtype=bool)
    assert np.issubdtype(data.dtype, bool)



from .compare import compare_datasets
from .conftest import base_config
import numpy as np
import openamundsen as oa
from pathlib import Path
import pytest
import tempfile


@pytest.fixture(scope='session')
def canopy_run():
    config = base_config()
    config.start_date = '2019-10-01'
    config.end_date = '2020-05-31'

    model = oa.OpenAmundsen(config)
    model.initialize()

    roi_xs = model.grid.X.flat[model.grid.roi_idxs_flat]
    roi_ys = model.grid.Y.flat[model.grid.roi_idxs_flat]

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_path = Path(temp_dir)

        for p in Path(config.input_data.grids.dir).glob('*.asc'):
            (tmp_path / p.name).symlink_to(p)

        config.input_data.grids.dir = str(temp_dir)

        soil = np.zeros(model.grid.shape, dtype=int)
        lc = np.zeros(model.grid.shape, dtype=int)

        soil[:] = 5
        lccs = model.config.land_cover.classes.keys()
        config.output_data.timeseries.points = []
        for lcc_num, lcc in enumerate(lccs):
            config.output_data.timeseries.points.append({
                'x': float(roi_xs[lcc_num]),
                'y': float(roi_ys[lcc_num]),
            })
            lc.flat[model.grid.roi_idxs_flat[lcc_num]] = lcc

        rio_meta = {'driver': 'AAIGrid'}
        oa.fileio.write_raster_file(
            oa.util.raster_filename('soil', config),
            soil.astype(np.int32),
            model.grid.transform,
            **rio_meta,
        )
        oa.fileio.write_raster_file(
            oa.util.raster_filename('lc', config),
            lc.astype(np.int32),
            model.grid.transform,
            **rio_meta,
        )

        config.canopy.enabled = True
        config.output_data.timeseries.add_default_points = False
        config.output_data.timeseries.variables = [
            {'var': 'snow.canopy_intercepted_load'},
            {'var': 'snow.canopy_intercepted_snowfall'},
            {'var': 'snow.canopy_sublimation'},
            {'var': 'snow.canopy_melt'},
            {'var': 'meteo.top_canopy_temp'},
        ]

        model = oa.OpenAmundsen(config)
        model.initialize()
        model.run()

    return model


@pytest.mark.slow
def test_canopy_snow(canopy_run):
    model = canopy_run
    ds = model.point_output.data
    lccs = model.config.land_cover.classes.keys()

    for lcc_num, lcc in enumerate(lccs):
        lcc_params = model.config.land_cover.classes[lcc]
        is_forest = lcc_params.get('is_forest', False)

        ds_lcc = ds.isel(point=lcc_num)

        if is_forest:
            assert np.all(ds_lcc.canopy_intercepted_load >= 0.)
            assert np.all(ds_lcc.canopy_intercepted_snowfall >= 0.)
            assert np.all(ds_lcc.canopy_sublimation >= 0.)
            assert np.all(ds_lcc.canopy_melt >= 0.)
            assert ds_lcc.canopy_intercepted_load.max() > 0.
            assert ds_lcc.canopy_intercepted_snowfall.max() > 0.
            assert ds_lcc.canopy_sublimation.max() > 0.
            assert ds_lcc.canopy_melt.max() > 0.
        else:
            assert np.all(np.isnan(ds_lcc.canopy_intercepted_load))
            assert np.all(np.isnan(ds_lcc.canopy_intercepted_snowfall))
            assert np.all(np.isnan(ds_lcc.canopy_sublimation))
            assert np.all(np.isnan(ds_lcc.canopy_melt))


@pytest.mark.slow
def test_canopy_meteorology(canopy_run):
    model = canopy_run
    ds = model.point_output.data
    lccs = model.config.land_cover.classes.keys()

    for lcc_num, lcc in enumerate(lccs):
        lcc_params = model.config.land_cover.classes[lcc]
        is_forest = lcc_params.get('is_forest', False)

        if is_forest:
            ds_lcc = ds.isel(point=lcc_num)
            assert np.all(ds_lcc.rel_hum >= 0.)
            assert np.all(ds_lcc.rel_hum <= 100.)
            assert np.all(ds_lcc.wind_speed >= 0.)
            assert np.all(ds_lcc.sw_in >= 0.)
            assert np.all(ds_lcc.lw_in >= 0.)
            assert np.all(np.abs(ds_lcc.temp - ds_lcc.top_canopy_temp) <= 10.)


def test_no_forest():
    config = base_config()
    config.start_date = '2019-10-01'
    config.end_date = '2019-10-01'

    model = oa.OpenAmundsen(config)
    model.initialize()

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_path = Path(temp_dir)

        for p in Path(config.input_data.grids.dir).glob('*.asc'):
            (tmp_path / p.name).symlink_to(p)

        config.input_data.grids.dir = str(temp_dir)

        lc = np.zeros(model.grid.shape, dtype=int)
        rio_meta = {'driver': 'AAIGrid'}
        oa.fileio.write_raster_file(
            oa.util.raster_filename('lc', config),
            lc.astype(np.int32),
            model.grid.transform,
            **rio_meta,
        )

        config.canopy.enabled = True
        model = oa.OpenAmundsen(config)
        model.initialize()
        model.run()


@pytest.mark.slow
@pytest.mark.comparison
def test_compare_canopy(canopy_run):
    compare_datasets(
        'canopy_point',
        canopy_run.point_output.data,
        point='point5',  # coniferous forest
    )



from loguru import logger
import numpy as np
from openamundsen import constants, interpolation, meteo, util


def _param_station_data(ds, param, date):
    """
    Return station measurements including station x, y and z coordinates for one or more given
    parameter(s) and a specified date.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the station measurements (i.e. the model.meteo
        variable)

    param : str or list
        Parameter(s) for which the measurements should be returned.

    date : datetime-like
        Date for which the measurements should be returned.

    Returns
    -------
    data : ndarray
        Measurements (excluding nodata values). If param is a string, this is a 1-D array; if param
        is a list of strings it is a 2-D array with dimensions
        (len(param), <number of available measurements where all parameters are non-null>).
    x, y, z : ndarrays
        Corresponding x, y and z coordinates.
    """
    if isinstance(param, str):
        param_as_list = [param]
    else:
        param_as_list = param

    ds_param = ds[param_as_list + ['x', 'y', 'alt']].sel(time=date).dropna(dim='station')
    data = ds_param[param]
    xs = ds_param['x'].values
    ys = ds_param['y'].values
    zs = ds_param['alt'].values

    if isinstance(param, str):
        data_vals = data.values
    else:
        data_vals = data.to_array().values

    return data_vals, xs, ys, zs


def _linear_fit(x, y):
    """
    Calculate the linear regression for a set of sample points.

    Parameters
    ----------
    x, y : ndarray
        x and y coordinates of the sample points.

    Returns
    -------
    slope, intercept : float
    """
    if len(x) < 2:
        slope = 0
        intercept = 0
    else:
        slope, intercept = np.polyfit(x, y, 1)

    return slope, intercept


def _apply_linear_trend(data, elevs, trend, direction):
    """
    Detrend or retrend a set of data points using a given trend value.

    Parameters
    ----------
    data : ndarray
        Detrended values.

    elevs : ndarray
        Elevations corresponding to the data points.

    trend : float
        Trend value.

    direction : str
        Either "detrend" or "retrend".

    Returns
    -------
    data_trend : ndarray
        De-/retrended data points.
    """
    if direction == 'detrend':
        return data - trend * elevs
    elif direction == 'retrend':
        return data + trend * elevs
    else:
        raise NotImplementedError(f'Unsupported direction: {direction}')


def _interpolate_with_trend(
    data,
    xs,
    ys,
    zs,
    target_xs,
    target_ys,
    target_zs,
    trend_method,
    lapse_rate,
):
    """
    Interpolate station measurements using inverse distance weighting with an
    elevation-dependent trend.

    Parameters
    ----------
    data : ndarray
        Values to be interpolated.

    xs, ys, zs : ndarray
        x, y and z coordinates of the stations.

    target_xs, target_ys, target_zs : ndarray
        x, y and z coordinates of the interpolation targets.

    trend_method : string
        Method to use for de-/retrending. Either 'regression' (to derive a
        trend from the data points using linear regression), 'fixed' (to use a
        prescribed lapse rate as linear trend), 'fractional' (to use a
        prescribed fractional lapse rate (in case of precipitation) as linear
        trend) or 'adjustment_factor' (to use a nonlinear precipitation
        adjustment factor following [1]).

    lapse_rate : float
        Trend value. If trend_method == 'regression', this value is ignored.
        If trend_method == 'fixed', the value is interpreted as absolute change
        by elevation (e.g. °C m-1).  If trend_method == 'fractional', the value
        is interpreted as a fractional change by elevation (e.g., 0.0005 for
        0.05 % m-1) If trend_method == 'adjustment_factor', the value is
        interpreted as a nonlinear precipitation adjustment factor following
        [1].

    Returns
    -------
    data_interpol : ndarray
        Interpolated values for the target locations.

    References
    ----------
    .. [1] Liston, G. E., & Elder, K. (2006). A Meteorological Distribution
       System for High-Resolution Terrestrial Modeling (MicroMet). Journal of
       Hydrometeorology, 7(2), 217–234. https://doi.org/10.1175/JHM486.1
    """
    if trend_method in ('regression', 'fixed', 'fractional'):
        if trend_method == 'regression':
            # When using the regression method, the passed lapse rate is overwritten
            lapse_rate, _ = _linear_fit(zs, data)
        elif trend_method == 'fixed':
            pass  # do nothing, i.e., use the passed lapse rate as is
        elif trend_method == 'fractional':
            lapse_rate *= np.nanmean(data)

        data_detrended = _apply_linear_trend(data, zs, lapse_rate, 'detrend')
        data_detrended_interpol = interpolation.idw(
            xs,
            ys,
            data_detrended,
            target_xs,
            target_ys,
        )
        data_interpol = _apply_linear_trend(data_detrended_interpol, target_zs, lapse_rate, 'retrend')
    elif trend_method == 'adjustment_factor':
        data_interpol_notrend = interpolation.idw(xs, ys, data, target_xs, target_ys)
        zs_interpol = interpolation.idw(xs, ys, zs, target_xs, target_ys)
        z_diffs = target_zs - zs_interpol
        data_interpol = data_interpol_notrend * (  # eq. (33) from [1]
            (1 + lapse_rate * z_diffs) / (1 - lapse_rate * z_diffs)
        )

    return data_interpol


def interpolate_station_data(model):
    """
    Interpolate station measurements to the model grid.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.
    """
    logger.debug('Interpolating station data')

    date = model.date
    roi = model.grid.roi
    target_xs = model.grid.roi_points[:, 0]
    target_ys = model.grid.roi_points[:, 1]
    target_zs = model.state.base.dem[roi]

    for param in ('temp', 'precip'):
        param_config = model.config['meteo']['interpolation'][constants.INTERPOLATION_CONFIG_PARAM_MAPPINGS[param]]
        data, xs, ys, zs = _param_station_data(model.meteo, param, date)

        model.state.meteo[param][roi] = interpolate_param(
            param,
            model.date,
            param_config,
            data,
            xs,
            ys,
            zs,
            target_xs,
            target_ys,
            target_zs,
        )

    # For relative humidity dew point temperature is interpolated, so we also need station
    # temperatures and the interpolated temperature field
    param = 'rel_hum'
    param_config = model.config['meteo']['interpolation'][constants.INTERPOLATION_CONFIG_PARAM_MAPPINGS[param]]
    data, xs, ys, zs = _param_station_data(model.meteo, [param, 'temp'], date)
    rel_hums = data[0, :]
    temps = data[1, :]
    model.state.meteo[param][roi] = interpolate_param(
        param,
        model.date,
        param_config,
        rel_hums,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
        temps=temps,
        target_temps=model.state.meteo['temp'][roi],
    )

    wind_config = model.config['meteo']['interpolation']['wind']
    if wind_config['method'] == 'idw':
        data, xs, ys, zs = _param_station_data(model.meteo, 'wind_speed', date)
        model.state.meteo['wind_speed'][roi] = interpolate_param(
            'wind_speed',
            model.date,
            wind_config,
            data,
            xs,
            ys,
            zs,
            target_xs,
            target_ys,
            target_zs,
        )

        if model._has_wind_gusts:
            data, xs, ys, zs = _param_station_data(model.meteo, 'wind_speed_gust', date)
            model.state.meteo['wind_speed_gust'][roi] = interpolate_param(
                'wind_speed',
                model.date,
                wind_config,
                data,
                xs,
                ys,
                zs,
                target_xs,
                target_ys,
                target_zs,
            )
    elif wind_config['method'] == 'liston':
        data, xs, ys, zs = _param_station_data(model.meteo, ['wind_speed', 'wind_dir'], date)
        wind_speeds = data[0, :]
        wind_dirs = data[1, :]
        wind_speed_corr, wind_dir_corr = _liston_wind_correction(
            model.date,
            wind_config,
            wind_speeds,
            wind_dirs,
            xs,
            ys,
            zs,
            target_xs,
            target_ys,
            target_zs,
            model.state.base.slope[roi],
            model.state.base.aspect[roi],
            model.state.base.scaled_curvature[roi],
        )
        model.state.meteo.wind_speed[roi] = np.clip(
            wind_speed_corr,
            *constants.ALLOWED_METEO_VAR_RANGES['wind_speed'],
        )
        model.state.meteo.wind_dir[roi] = wind_dir_corr

        if model._has_wind_gusts:
            data, xs, ys, zs = _param_station_data(model.meteo, ['wind_speed_gust', 'wind_dir'], date)
            wind_speed_gusts = data[0, :]
            wind_dirs = data[1, :]
            wind_speed_gusts_corr, _ = _liston_wind_correction(
                model.date,
                wind_config,
                wind_speed_gusts,
                wind_dirs,
                xs,
                ys,
                zs,
                target_xs,
                target_ys,
                target_zs,
                model.state.base.slope[roi],
                model.state.base.aspect[roi],
                model.state.base.scaled_curvature[roi],
            )
            model.state.meteo.wind_speed_gust[roi] = np.clip(
                wind_speed_gusts_corr,
                *constants.ALLOWED_METEO_VAR_RANGES['wind_speed_gust'],
            )
    else:
        raise NotImplementedError(f'Unsupported method: {wind_config.method}')


def interpolate_param(
        param,
        date,
        param_config,
        data,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
        temps=None,
        target_temps=None,
):
    """
    Interpolate a set of data points to a set of target points.

    Parameters
    ----------
    param : str
        Parameter to be interpolated. Can be one of 'temp', 'precip', 'rel_hum', 'wind_speed'.

    date : datetime-like
        Date (required in the case of seasonally specified lapse rates).

    param_config : dict
        Interpolation configuration for the respective parameter from the model run configuration
        (e.g., config['meteo']['interpolation'][‘temperature']).

    data : ndarray
        Values to be interpolated.

    xs, ys, zs : ndarray
        x, y and z coordinates of the stations.

    target_xs, target_ys, target_zs : ndarray
        x, y and z coordinates of the interpolation targets.

    temps : ndarray, default None
        Temperatures at the source locations (required for humidity interpolation).

    target_temps : ndarray, default None
        Temperatures at the target locations (required for humidity interpolation).

    Returns
    -------
    data_interpol : ndarray
        Interpolated values for the target locations.
    """
    # If there are no points to be interpolated, return an all-zero array for precipitation, and an
    # all-nan array for all other parameters
    if data.size == 0:
        if param == 'precip':
            fill_value = 0.
        else:
            fill_value = np.nan

        return np.full(target_xs.shape, fill_value)

    if param in ('temp', 'precip', 'rel_hum', 'wind_speed', 'wind_vec'):
        trend_method = param_config['trend_method']
        lapse_rate = param_config['lapse_rate'][date.month - 1]
    else:
        raise NotImplementedError(f'Unsupported parameter: {param}')

    # For relative humidity interpolate dew point temperature and convert back to humidity later
    if param == 'rel_hum':
        if temps is None or target_temps is None:
            raise Exception('Temperature must be provided for humidity interpolation')

        rel_hums = data
        dew_point_temps = meteo.dew_point_temperature(temps, rel_hums)
        data = dew_point_temps

    data_interpol = _interpolate_with_trend(
        data,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
        trend_method,
        lapse_rate,
    )

    if param == 'rel_hum':
        target_dew_point_temps = data_interpol
        vapor_press = meteo.saturation_vapor_pressure(target_dew_point_temps, 'water')
        sat_vapor_press = meteo.saturation_vapor_pressure(target_temps, 'water')
        data_interpol = 100 * vapor_press / sat_vapor_press

    # Restrict interpolated values to the range of the point values if extrapolation is disabled in
    # the parameter config
    if not param_config['extrapolate'] and len(data) > 0:
        min_range = np.nanmin(data)
        max_range = np.nanmax(data)

        if np.isfinite(min_range):
            data_interpol = np.clip(data_interpol, min_range, max_range)

    if param in constants.ALLOWED_METEO_VAR_RANGES:
        min_range, max_range = constants.ALLOWED_METEO_VAR_RANGES[param]
        data_interpol = np.clip(data_interpol, min_range, max_range)

    return data_interpol


def _liston_wind_correction(
    date,
    wind_config,
    wind_speeds,
    wind_dirs,
    xs,
    ys,
    zs,
    target_xs,
    target_ys,
    target_zs,
    target_slopes,
    target_aspects,
    target_scaled_curvatures,
):
    wind_us, wind_vs = meteo.wind_to_uv(wind_speeds, wind_dirs)
    wind_u_roi = interpolate_param(
        'wind_vec',
        date,
        wind_config,
        wind_us,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
    )
    wind_v_roi = interpolate_param(
        'wind_vec',
        date,
        wind_config,
        wind_vs,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
    )
    wind_speed_roi, wind_dir_roi = meteo.wind_from_uv(wind_u_roi, wind_v_roi)
    wind_dir_minus_aspect = np.deg2rad(wind_dir_roi - target_aspects)
    wind_slope = (  # slope in the direction of the wind (eq. (15))
        np.deg2rad(target_slopes) * np.cos(wind_dir_minus_aspect)
    )
    wind_slope_scaled = util.normalize_array(wind_slope, -0.5, 0.5)
    wind_weighting_factor = (  # eq. (16)
        1
        + wind_config.slope_weight * wind_slope_scaled
        + wind_config.curvature_weight * target_scaled_curvatures
    )
    wind_dir_diverting_factor = (  # eq. (18)
        -0.5 * wind_slope_scaled * np.sin(-2 * wind_dir_minus_aspect)
    )
    wind_speed_corr = wind_speed_roi * wind_weighting_factor
    wind_dir_corr = (wind_dir_roi + wind_dir_diverting_factor) % 360

    return wind_speed_corr, wind_dir_corr



from .meteo import (
    read_csv_meteo_file,
    read_meteo_data,
    read_netcdf_meteo_file,
)

from .griddedoutput import GriddedOutputManager
from .pointoutput import PointOutputManager

from .raster import (
    read_raster_metadata,
    read_raster_file,
    write_raster_file,
)



from numba import njit


@njit(cache=True)
def solve_tridiag(a, b, c, d, overwrite_bd=False):
    """
    Solve a tridiagonal equation system using the Thomas algorithm.

    The equivalent using scipy.linalg.solve_banded would be:

        ab = np.zeros((3, len(a)))
        ab[0, 1:] = c[:-1]
        ab[1, :] = b
        ab[2, :-1] = a[1:]
        x = scipy.linalg.solve_banded((1, 1), ab, d)

    Parameters
    ----------
    a : ndarray
        Lower diagonal of the tridiagonal matrix as a length n array with
        elements (0, a_2, a_3, ..., a_n).

    b : ndarray
        Main diagonal of the tridiagonal matrix as a length n array with
        elements (b_1, ..., b_n).

    c : ndarray
        Upper diagonal of the tridiagonal matrix as a length n array with
        elements (c_1, ..., c_n-1, 0).

    d : ndarray
        Right hand side of the matrix system as a length n array.

    Returns
    -------
    x : ndarray
        Solution vector as a length n array.
    """
    n = len(d)  # number of equations

    if not overwrite_bd:
        b = b.copy()
        d = d.copy()

    for k in range(1, n):
        m = a[k] / b[k - 1]
        b[k] = b[k] - m * c[k - 1]
        d[k] = d[k] - m * d[k - 1]

    x = b
    x[-1] = d[-1] / b[-1]

    for k in range(n - 2, 0 - 1, -1):
        x[k] = (d[k] - c[k] * x[k + 1]) / b[k]

    return x


def solve_tridiag_array(a, b, c, d, overwrite_bd=False):
    """
    Solve multiple tridiagonal equation systems using the Thomas algorithm.
    Parameters are same as for solve_tridiag(), however are in this case arrays
    with dimensions (n, num_equations).
    """
    n = d.shape[0]  # number of equations

    if not overwrite_bd:
        b = b.copy()
        d = d.copy()

    for k in range(1, n):
        m = a[k, :] / b[k - 1, :]
        b[k, :] = b[k, :] - m * c[k - 1, :]
        d[k, :] = d[k, :] - m * d[k - 1, :]

    x = b
    x[-1, :] = d[-1, :] / b[-1, :]

    for k in range(n - 2, 0 - 1, -1):
        x[k, :] = (d[k, :] - c[k, :] * x[k + 1, :]) / b[k, :]

    return x



from .conftest import base_config
from openamundsen.util import to_yaml
import os
import pytest
import subprocess
import textwrap
import xarray as xr


def test_numba(tmp_path):
    config = base_config()
    config.start_date = '2020-01-18'
    config.end_date = '2020-01-18'
    config.results_dir = tmp_path
    config.output_data.timeseries.format = 'netcdf'
    config.output_data.grids.format = 'netcdf'
    config.output_data.grids.variables = [
        {'var': 'meteo.temp'},
        {'var': 'snow.swe'},
        {'var': 'snow.temp', 'name': 'snow_temp'},
    ]

    config_file = tmp_path / 'config.yml'
    with open(config_file, 'w') as f:
        f.write(to_yaml(config.toDict()))

    subprocess.check_call(['openamundsen', str(config_file)])
    ds_points_numba = xr.load_dataset(tmp_path / 'output_timeseries.nc')
    ds_grids_numba = xr.load_dataset(tmp_path / 'output_grids.nc')

    env = os.environ.copy()
    env['NUMBA_DISABLE_JIT'] = '1'
    subprocess.check_call(['openamundsen', str(config_file)], env=env)
    ds_points_nonumba = xr.load_dataset(tmp_path / 'output_timeseries.nc')
    ds_grids_nonumba = xr.load_dataset(tmp_path / 'output_grids.nc')

    assert ds_points_numba.equals(ds_points_nonumba)
    assert ds_grids_numba.equals(ds_grids_nonumba)


@pytest.mark.slow
def test_floating_errors(tmp_path):
    config = base_config()
    config.results_dir = tmp_path

    config_file = tmp_path / 'config.yml'
    with open(config_file, 'w') as f:
        f.write(to_yaml(config.toDict()))

    with open(tmp_path / 'run.py', 'w') as f:
        f.write(textwrap.dedent(
            f'''
            import numpy as np
            np.seterr(all='raise')

            import openamundsen as oa

            config = oa.read_config('{config_file}')
            model = oa.OpenAmundsen(config)
            model.initialize()
            model.run()
            '''
        ))

    env = os.environ.copy()
    env['NUMBA_DISABLE_JIT'] = '1'
    subprocess.check_call(['python', tmp_path / 'run.py'], env=env)



import argparse
import openamundsen as oa


def main():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('config_file', help='configuration file')
    args = parser.parse_args()

    config = oa.read_config(args.config_file)
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()



import numpy as np
from numpy.testing import assert_allclose
from openamundsen.tridiag import solve_tridiag
from scipy.sparse import diags


def test_solve_tridiag():
    N = 100

    a = np.append([0], np.random.rand(N - 1))
    b = np.random.rand(N)
    c = np.append(np.random.rand(N - 1), [0])
    x = np.random.rand(N)

    A = diags([a[1:], b, c[:-1]], [-1, 0, 1])
    d = A @ x

    x_solved = solve_tridiag(a, b, c, d)
    assert_allclose(x_solved, x)



import numpy as np
import openamundsen.constants as c
import openamundsen.modules.radiation as rad


def _clear_sky_shortwave_irradiance(
    day_angle,
    sun_vec,
    shadows,
    dem,
    svf,
    normal_vec,
    atmos_press,
    precipitable_water,
    ground_albedo,
    roi=None,
    ozone_layer_thickness=0.0035,
    atmospheric_visibility=25000.,
    single_scattering_albedo=0.9,
    clear_sky_albedo=0.0685,
):
    """
    Calculate potential direct and diffuse shortwave irradiance for a given DEM
    and sun position after Corripio (2002).

    Parameters
    ----------
    day_angle : float
        Day angle (radians).

    sun_vec : array-like
        Solar vector.

    shadows : ndarray
        Array containing 1 for illuminated and 0 for shadowed areas.

    dem : ndarray
        Elevation (m).

    svf : ndarray
        Sky view factor.

    normal_vec : ndarray with dimensions (3, *dem.shape)
        Unit vector perpendicular to the surface.

    atmos_press : ndarray
        Atmospheric pressure (Pa).

    precipitable_water : ndarray
        Precipitable water (kg m-2).

    ground_albedo : float
        Mean surface albedo.

    roi : ndarray, default None
        Boolean array specifying the region of interest.

    ozone_layer_thickness : float, default 0.0035
        Ozone layer thickness (m).

    visibility : float, default 25000
        Atmospheric visibility (m).

    single_scattering_albedo : float, default 0.9
        Single scattering albedo.

    clear_sky_albedo : float, default 0.0685
        Clear sky albedo.

    Returns
    -------
    dir_irr : ndarray
        Potential direct irradiance (W m-2).

    diff_irr : ndarray
        Potential diffuse irradiance (W m-2).

    References
    ----------
    .. [1] Corripio, J. G. (2002). Modelling the energy balance of high
       altitude glacierised basins in the Central Andes. PhD thesis, University of
       Edinburgh.
    """
    zenith_angle = np.arccos(sun_vec[2])

    if roi is None:
        roi = np.ones(dem.shape, dtype=bool)

    # Eccentricity correction (Corripio, 2002)
    ecc_corr = (
        1.000110
        + 0.034221 * np.cos(day_angle)
        + 0.001280 * np.sin(day_angle)
        + 0.000719 * np.cos(2 * day_angle)
        + 0.000077 * np.sin(2 * day_angle)
    )

    # Angle of incidence of the sun on inclined surfaces (dot product of the
    # unit vector normal to the surface and the unit vector in the direction
    # of the sun)
    aoi = (
        sun_vec[0] * normal_vec[0, :] +
        sun_vec[1] * (-normal_vec[1, :]) +
        sun_vec[2] * normal_vec[2, :]
    )
    aoi = aoi.clip(min=0)  # shadows due to steepness of the slope (sun "behind")

    # Relative optical air mass for standard pressure (Iqbal 1983)
    rel_opt_air_mass = 1. / (
        np.cos(zenith_angle)
        + 0.15 * (93.885 - np.rad2deg(zenith_angle))**(-1.253)
    ).clip(max=14.1952)  # function is only valid for zenith angles up to 87°

    # Relative optical air mass pressure corrected
    rel_opt_air_mass_press_corr = rel_opt_air_mass * atmos_press / c.STANDARD_ATMOSPHERE

    ts = _transmittances(
        dem,
        atmos_press,
        precipitable_water,
        rel_opt_air_mass,
        ozone_layer_thickness,
        atmospheric_visibility,
        single_scattering_albedo,
    )

    trans_dir = (
        ts['trans_rayleigh']
        * ts['trans_ozone']
        * ts['trans_gases']
        * ts['trans_vapor']
        * ts['trans_aerosols']
        + ts['elev_corr']
    )
    trans_dir = np.clip(trans_dir, 0, 1)

    top_atmosphere_rad = c.SOLAR_CONSTANT * ecc_corr  # top of atmosphere radiation
    pot_irr = top_atmosphere_rad * aoi * shadows  # potential irradiance including shadows

    dir_irr = 0.9751 * pot_irr * trans_dir

    # Rayleigh-scattered diffuse irradiance
    diff_irr_rayleigh = (
        0.79
        * top_atmosphere_rad * np.cos(zenith_angle)
        * ts['trans_ozone']
        * ts['trans_gases']
        * ts['trans_vapor']
        * ts['trans_aerosol_abs']
        * 0.5 * (1 - ts['trans_rayleigh'])
        / (1 - rel_opt_air_mass_press_corr + rel_opt_air_mass_press_corr**1.02)
    )

    # Aerosol-scattered diffuse irradiance
    scattering_ratio = (
        -0.2562 * zenith_angle**2
        * 0.1409 * zenith_angle + 0.9067
    )
    trans_aerosol_ratio = ts['trans_aerosols'] / ts['trans_aerosol_abs']
    diff_irr_aerosols = (
        0.79
        * top_atmosphere_rad * np.cos(zenith_angle)
        * ts['trans_ozone']
        * ts['trans_gases']
        * ts['trans_vapor']
        * ts['trans_aerosol_abs']
        * scattering_ratio
        * (1 - trans_aerosol_ratio)
        / (1 - rel_opt_air_mass_press_corr + rel_opt_air_mass_press_corr**1.02)
    )

    # Multiply-reflected irradiance between the earth and the atmosphere
    atmospheric_albedo = (
        clear_sky_albedo
        + (1 - scattering_ratio)
        * (1 - trans_aerosol_ratio)
    )
    diff_irr_atmos_ref = (
        (
            dir_irr * np.cos(zenith_angle)
            + diff_irr_rayleigh
            + diff_irr_aerosols
        )
        * ground_albedo * atmospheric_albedo
        / (1 - ground_albedo * atmospheric_albedo)
    )

    diff_irr = diff_irr_rayleigh + diff_irr_aerosols + diff_irr_atmos_ref

    # Diffuse reflected radiation from surrounding slopes
    sun_frac = shadows[roi].mean()
    pos_sun = shadows[roi] == 1
    if pos_sun.sum() > 0:
        mean_sun_aoi = np.nanmean(aoi[roi][pos_sun])
    else:
        mean_sun_aoi = 0.
    diff_irr_terrain_ref = (
        (1 - svf) * (
            pot_irr
            * trans_dir
            * sun_frac
            * mean_sun_aoi
            + diff_irr
        ) * ground_albedo
    )

    # Diffuse irradiance corrected for topography
    diff_irr_topo_corr = diff_irr * svf + diff_irr_terrain_ref

    return dir_irr, diff_irr_topo_corr


def _transmittances(
    elev,
    atmos_press,
    precipitable_water,
    rel_opt_air_mass,
    ozone_layer_thickness,
    visibility,
    single_scattering_albedo,
):
    """
    Calculate atmospheric transmittances after Corripio (2002).

    Parameters
    ----------
    elev : ndarray
        Elevation (m).

    atmos_press : ndarray
        Atmospheric pressure (Pa).

    precipitable_water : ndarray
        Precipitable water (kg m-2).

    rel_opt_air_mass : ndarray
        Relative optical air mass for standard pressure.

    ozone_layer_thickness : float
        Ozone layer thickness (m).

    visibility : float
        Atmospheric visibility (m).

    single_scattering_albedo : float
        Single scattering albedo.

    Returns
    -------
    transmittances : dict
        Dictionary containing the following fields:
        - 'trans_rayleigh': Rayleigh scattering
        - 'trans_ozone': transmittance by ozone
        - 'trans_gases': transmittance by uniformy mixed gases
        - 'trans_vapor': transmittance by water vapor
        - 'trans_aerosols': transmittance by aerosols
        - 'trans_aerosol_abs': transmittance of direct radiation due to aerosol
          absorptance
        - 'elev_corr': correction of atmospheric transmittances for altitude
          after Bintanja (1996)

    References
    ----------
    .. [1] Corripio, J. G. (2002). Modelling the energy balance of high
       altitude glacierised basins in the Central Andes. PhD thesis, University of
       Edinburgh.
    """
    ozone_layer_thickness_cm = ozone_layer_thickness * 100.
    visibility_km = visibility / 1000.

    # Relative optical air mass pressure corrected (eq. (3.10))
    rel_opt_air_mass_press_corr = rel_opt_air_mass * atmos_press / c.STANDARD_ATMOSPHERE

    # Rayleigh scattering
    trans_rayleigh = (
        np.exp(-0.0903 * rel_opt_air_mass_press_corr**0.84)
        * (1 + rel_opt_air_mass_press_corr - rel_opt_air_mass_press_corr**1.01)
    )

    # Transmittance by ozone (constant for the entire area) (eq. (3.13))
    lm = ozone_layer_thickness_cm * rel_opt_air_mass
    trans_ozone = (
        1 - (
            0.1611 * lm * (1 + 139.48 * lm)**(-0.3035)  # -0.3035 seems to be the correct value (in eq. (3.13) in [1] the value is -0.035)
            - 0.002715 * lm * 1. / (1 + 0.044 * lm + 3e-4 * lm**2)
        )
    )

    # Transmittance by uniformly mixed gases (eq. (3.15))
    trans_gases = np.exp(-0.0127 * rel_opt_air_mass_press_corr**0.26)

    # Transmittance by water vapor (eq. (3.16))
    # Note: eq. (3.16) in [1] is missing a closing parenthesis; the exponent 0.6828 should apply to
    # the entire (1 + 79.034 * w * m_r) term, see also e.g. eq. (13) in
    # https://doi.org/10.1029/2003JD003973
    precipitable_water_gcm2 = precipitable_water / 10  # kg m-2 (= mm) to g cm-2 (= cm)
    wm = precipitable_water_gcm2 * rel_opt_air_mass
    trans_vapor = (
        1
        - 2.4959 * wm
        * 1. / ((1 + 79.034 * wm)**0.6828 + 6.385 * wm)
    )

    # Transmittance by aerosols (eq. (3.17))
    trans_aerosols = (0.97 - 1.265 * visibility_km**(-0.66))**(rel_opt_air_mass_press_corr**0.9)

    # Correction of atmospheric transmittances for altitude (eq. (3.8))
    elev_corr = 2.2e-5 * elev.clip(max=3000)

    # Transmittance of direct radiation due to aerosol absorptance (eq. (3.19))
    trans_aerosol_abs = (
        1
        - (1 - single_scattering_albedo)
        * (1 - rel_opt_air_mass_press_corr + rel_opt_air_mass_press_corr**1.06)
        * (1 - trans_aerosols)
    )

    return {
        'trans_rayleigh': trans_rayleigh,
        'trans_ozone': trans_ozone,
        'trans_gases': trans_gases,
        'trans_vapor': trans_vapor,
        'trans_aerosols': trans_aerosols,
        'trans_aerosol_abs': trans_aerosol_abs,
        'elev_corr': elev_corr,
    }



from .conftest import base_config
from openamundsen.util import to_yaml
import subprocess


def test_cli(tmp_path):
    config = base_config()
    config.start_date = '2020-01-18 00:00'
    config.end_date = '2020-01-18 00:00'
    config.results_dir = tmp_path

    config_file = tmp_path / 'config.yml'
    with open(config_file, 'w') as f:
        f.write(to_yaml(config.toDict()))

    subprocess.check_call(['openamundsen', str(config_file)])



from .conftest import base_config
import numpy as np
from numpy.testing import assert_allclose
import openamundsen as oa
import openamundsen.errors as errors
from openamundsen.fileio.griddedoutput import _freq_write_dates
import pandas as pd
import pytest
import rasterio
import xarray as xr


def test_freq_write_dates():
    dates = pd.date_range(start='2021-01-01 00:00', end='2021-12-31 23:00', freq='h')

    assert _freq_write_dates(dates, 'h', False).equals(dates)

    wd = _freq_write_dates(dates, '3h', False)
    assert np.all(wd.hour.isin([0, 3, 6, 9, 12, 15, 18, 21]))

    wd = _freq_write_dates(dates, 'D', False)
    assert np.all(wd.hour == 0)
    assert dates.normalize().unique().equals(wd.normalize())

    wd = _freq_write_dates(dates, 'D', True)
    assert np.all(wd.hour == 23)

    wd = _freq_write_dates(dates, 'ME', True)
    assert np.all(wd.day.isin([28, 30, 31]))
    assert np.all(wd.hour == 23)

    assert _freq_write_dates(dates, 'YE', False).equals(pd.to_datetime(['2021-12-31 23:00']))
    assert _freq_write_dates(dates, 'YS', False).equals(pd.to_datetime(['2021-01-01 00:00']))

    wd = _freq_write_dates(dates, 'ME', False)
    assert np.all(wd.day.isin([28, 30, 31]))
    assert np.all(wd.hour == 23)

    wd = _freq_write_dates(dates, 'MS', False)
    assert np.all(wd.day == 1)
    assert np.all(wd.hour == 0)

    dates = pd.date_range(start='2021-01-01 06:00', end='2021-12-31 21:00', freq='3h')
    wd = _freq_write_dates(dates, 'D', False)
    assert np.all(wd.hour == 6)

    wd = _freq_write_dates(dates, 'D', True)
    assert np.all(wd.hour == 3)

    with pytest.raises(ValueError):
        wd = _freq_write_dates(dates, 'h', False)

    with pytest.raises(ValueError):
        wd = _freq_write_dates(dates, 'h', True)

    dates = pd.date_range(start='2021-01-01 02:00', end='2021-12-31 04:00', freq='3h')
    wd = _freq_write_dates(dates, 'D', False)
    assert np.all(wd.hour == 2)
    # assert _freq_write_dates(dates, 'A', False).size == 0

    dates = pd.date_range(start='2021-01-01 02:00', end='2021-01-01 07:00', freq='h')
    assert len(_freq_write_dates(dates, 'D', False)) == 1
    assert len(_freq_write_dates(dates, 'D', True)) == 0

    dates = pd.date_range(start='2021-01-15 00:00', end='2021-03-15 21:00', freq='3h')
    wd = _freq_write_dates(dates, 'ME', True)
    assert len(wd) == 2

    dates = pd.date_range(start='2021-01-01 01:00', end='2021-12-31 22:00', freq='3h')
    assert _freq_write_dates(dates, 'YE', False).equals(pd.to_datetime(['2021-12-31 22:00']))
    assert _freq_write_dates(dates, 'YS', False).equals(pd.to_datetime(['2021-01-01 01:00']))

    for freq in ('BME', 'QE', 'ms', 'us', 'ns', 'foo'):
        with pytest.raises(errors.ConfigurationError):
            _freq_write_dates(dates, freq, False)


@pytest.mark.parametrize('fmt', ['netcdf', 'ascii', 'geotiff', 'memory'])
def test_formats(fmt, tmp_path):
    # Skip test for format "ascii" for certain GDAL versions because of this bug:
    # https://github.com/OSGeo/gdal/issues/9666
    if fmt == 'ascii' and rasterio.__gdal_version__ in (
        '3.8.0',
        '3.8.1',
        '3.8.2',
        '3.8.3',
        '3.8.4',
        '3.8.5',
    ):
        pytest.skip()

    config = base_config()
    config.end_date = '2020-01-16'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = fmt
    grid_cfg.variables = [
        {'var': 'meteo.temp', 'freq': 'D'},
        {'var': 'meteo.precip', 'freq': 'D', 'agg': 'sum'},
        {'var': 'meteo.sw_in', 'dates': ['2020-01-15 12:00'], 'name': 'myvar'},
        {'var': 'base.dem', 'dates': ['2020-01-15 00:00']},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    if fmt in ('netcdf', 'memory'):
        if fmt == 'netcdf':
            ds = xr.open_dataset(tmp_path / 'output_grids.nc')
        elif fmt == 'memory':
            ds = model.gridded_output.data

        assert 'temp' in ds.variables
        assert ds.time1.to_index().equals(pd.DatetimeIndex([
            '2020-01-15 00:00',
            '2020-01-16 00:00',
        ]))
        assert 'precip' in ds.variables
        assert ds.time2.to_index().equals(pd.DatetimeIndex([
            '2020-01-15 21:00',
            '2020-01-16 21:00',
        ]))
        assert pd.DatetimeIndex(ds.time2_bounds.values.ravel()).equals(pd.DatetimeIndex([
            '2020-01-15 00:00:00',
            '2020-01-15 21:00:00',
            '2020-01-15 21:00:00',
            '2020-01-16 21:00:00',
        ]))
        assert 'myvar' in ds.variables
        assert ds.time3.to_index().equals(pd.DatetimeIndex(['2020-01-15 12:00']))
        assert_allclose(model.state.base.dem, ds.dem.loc['2020-01-15 00:00', :, :].values)
    else:
        if fmt == 'ascii':
            ext = 'asc'
        elif fmt == 'geotiff':
            ext = 'tif'

        fn = tmp_path / f'temp_2020-01-16T0000.{ext}'
        with rasterio.open(fn) as ds:
            ds.read(1)

        fn = tmp_path / f'precip_2020-01-16T0000_2020-01-16T2100.{ext}'
        with rasterio.open(fn) as ds:
            ds.read(1)

        fn = tmp_path / f'myvar_2020-01-15T1200.{ext}'
        with rasterio.open(fn) as ds:
            ds.read(1)

        fn = tmp_path / f'dem_2020-01-15T0000.{ext}'
        with rasterio.open(fn) as ds:
            assert_allclose(model.state.base.dem, ds.read(1))


def test_duplicate_output_names(tmp_path):
    config = base_config()
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {'var': 'meteo.temp'},
        {'var': 'snow.temp'},
    ]

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()


def test_values(tmp_path):
    config = base_config()
    config.start_date = '2020-01-17'
    config.end_date = '2020-01-19'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {'var': 'meteo.temp', 'freq': 'D', 'agg': 'mean'},
        {'var': 'meteo.precip', 'freq': 'D', 'agg': 'sum'},
        {'var': 'meteo.sw_in', 'dates': ['2020-01-17 12:00']},
        {'var': 'snow.swe', 'freq': 'D'},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    data_vals = {
        'temp': {},
        'precip': {},
        'sw_in': {},
        'swe': {},
    }

    for date in model.dates:
        model.run_single()
        date = date.strftime('%Y-%m-%d %H:%M')
        data_vals['temp'][date] = model.state.meteo.temp.copy()
        data_vals['precip'][date] = model.state.meteo.precip.copy()
        data_vals['sw_in'][date] = model.state.meteo.sw_in.copy()
        data_vals['swe'][date] = model.state.snow.swe.copy()

    ds = model.gridded_output.data
    assert_allclose(
        data_vals['sw_in']['2020-01-17 12:00'],
        ds.sw_in.loc['2020-01-17 12:00', :, :].values,
    )

    assert_allclose(
        data_vals['swe']['2020-01-19 00:00'],
        ds.swe.loc['2020-01-19 00:00', :, :].values,
    )

    mean_temp = (
        data_vals['temp']['2020-01-18 00:00']
        + data_vals['temp']['2020-01-18 03:00']
        + data_vals['temp']['2020-01-18 06:00']
        + data_vals['temp']['2020-01-18 09:00']
        + data_vals['temp']['2020-01-18 12:00']
        + data_vals['temp']['2020-01-18 15:00']
        + data_vals['temp']['2020-01-18 18:00']
        + data_vals['temp']['2020-01-18 21:00']
    ) / 8
    assert_allclose(
        mean_temp,
        ds.temp.loc['2020-01-18 21:00', :, :].values,
    )

    precip_sum = (
        data_vals['precip']['2020-01-18 00:00']
        + data_vals['precip']['2020-01-18 03:00']
        + data_vals['precip']['2020-01-18 06:00']
        + data_vals['precip']['2020-01-18 09:00']
        + data_vals['precip']['2020-01-18 12:00']
        + data_vals['precip']['2020-01-18 15:00']
        + data_vals['precip']['2020-01-18 18:00']
        + data_vals['precip']['2020-01-18 21:00']
    )
    assert_allclose(
        precip_sum,
        ds.precip.loc['2020-01-18 21:00', :, :].values,
    )


def test_data_type(tmp_path):
    config = base_config()
    config.start_date = '2020-01-17'
    config.end_date = '2020-01-17'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {'var': 'meteo.temp', 'freq': 'D', 'agg': 'mean'},
        {'var': 'snow.num_layers', 'freq': 'D'},
        {'var': 'snow.num_layers', 'freq': 'D', 'agg': 'sum', 'name': 'num_layers_sum'},
        {'var': 'snow.num_layers', 'freq': 'D', 'agg': 'mean', 'name': 'num_layers_mean'},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    ds = model.gridded_output.data
    assert np.issubdtype(ds.temp.dtype, np.float32)
    assert np.issubdtype(ds.num_layers.dtype, np.integer)
    assert np.issubdtype(ds.num_layers_sum.dtype, np.integer)
    assert np.issubdtype(ds.num_layers_mean.dtype, np.float32)


def test_nothing_to_write(tmp_path):
    config = base_config()
    config.start_date = '2020-01-01'
    config.end_date = '2020-01-01 03:00'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.variables = [
        {'var': 'snow.swe', 'freq': 'D', 'agg': 'sum'},
        {'var': 'meteo.temp', 'dates': ['2020-01-02 12:00']},
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    assert model.gridded_output.data is None

    grid_cfg.variables = [
        {'var': 'snow.swe', 'freq': 'D', 'agg': 'sum'},
        {'var': 'meteo.temp', 'dates': ['2020-01-01 02:00']},
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.gridded_output.data
    assert list(ds.data_vars.keys()) == ['temp']


def test_dask(tmp_path):
    config = base_config()
    config.end_date = '2020-01-16'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = 'netcdf'
    grid_cfg.variables = [
        {'var': 'meteo.temp', 'freq': 'D'},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()

    ds_inmem = model.gridded_output._create_dataset(in_memory=True)
    ds_dask = model.gridded_output._create_dataset(in_memory=False)
    assert ds_inmem.temp.chunks is None
    assert ds_dask.temp.chunks is not None


def test_append(tmp_path):
    config = base_config()
    config.start_date = '2020-01-17'
    config.end_date = '2020-01-19'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = 'netcdf'
    grid_cfg.variables = [
        {'var': 'meteo.rel_hum'},
        {'var': 'meteo.temp', 'freq': 'D', 'agg': 'mean'},
        {'var': 'meteo.precip', 'freq': 'D', 'agg': 'sum'},
        {'var': 'meteo.sw_in', 'dates': ['2020-01-17 12:00']},
        {'var': 'snow.swe', 'freq': 'D'},
    ]

    grid_cfg.append = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    for date_num, date in enumerate(model.dates):
        model.run_single()
        if date_num == 0:
            ds = xr.load_dataset(tmp_path / 'output_grids.nc')
            assert ds.sizes['time1'] == len(model.dates)
    ds_no_append = xr.load_dataset(tmp_path / 'output_grids.nc')

    grid_cfg.append = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    for date_num, date in enumerate(model.dates):
        model.run_single()
        with xr.open_dataset(tmp_path / 'output_grids.nc') as ds:
            assert ds.sizes['time1'] == date_num + 1
    ds_append = xr.load_dataset(tmp_path / 'output_grids.nc')

    xr.testing.assert_identical(ds_no_append, ds_append)


@pytest.mark.parametrize('fmt', ['netcdf', 'geotiff'])
def test_compress(fmt, tmp_path):
    config = base_config()
    config.start_date = '2020-01-01 00:00'
    config.end_date = '2020-01-01 03:00'
    config.results_dir = tmp_path
    grid_cfg = config.output_data.grids
    grid_cfg.format = fmt
    grid_cfg.variables = [
        {'var': 'meteo.temp'},
    ]

    for compress in (False, True):
        grid_cfg.compress = compress
        model = oa.OpenAmundsen(config)
        model.initialize()
        model.run()

        if fmt == 'netcdf':
            with xr.open_dataset(tmp_path / 'output_grids.nc') as ds:
                assert ds['temp'].encoding['zlib'] is compress
        elif fmt == 'geotiff':
            with rasterio.open(tmp_path / 'temp_2020-01-01T0000.tif') as ds:
                if compress:
                    assert ds.compression is not None
                else:
                    assert ds.compression is None



class YamlReaderError(Exception):
    pass


class ConfigurationError(Exception):
    pass


class RasterFileError(Exception):
    pass


class MeteoDataError(Exception):
    pass


class CategoryError(Exception):
    pass


class RuntimeError(Exception):
    pass



from .conftest import base_config
import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import openamundsen as oa
from openamundsen import errors, meteo as oameteo
import pandas as pd
from pathlib import Path
import pytest
import xarray as xr


def meteo_to_netcdf(ds, out_dir):
    ds = ds.copy(deep=True)
    td = ds.time[1] - ds.time[0]
    timestep = td.values / np.timedelta64(1, 's')

    ds = ds[[
        'station_name',
        'lon',
        'lat',
        'alt',
        'time',
        'temp',
        'precip',
        'rel_hum',
        'sw_in',
        'wind_speed',
        'wind_dir',
    ]].rename_vars({
        'temp': 'tas',
        'precip': 'pr',
        'rel_hum': 'hurs',
        'sw_in': 'rsds',
        'wind_speed': 'wss',
        'wind_dir': 'wind_dir',
    })

    ds['pr'] /= timestep
    ds['pr'].attrs = {
        'standard_name': 'precipitation_flux',
        'units': 'kg m-2 s-1',
    }

    for station_num in range(len(ds.station)):
        ds_cur = ds.isel(station=station_num).copy()
        station_id = str(ds_cur.station.values)
        station_name = str(ds_cur.station_name.values)
        ds_cur.attrs = {
            'Conventions': 'CF-1.6',
            'station_name': station_name,
        }
        ds_cur = ds_cur.drop_vars(['station', 'station_name'])
        ds_cur.to_netcdf(f'{out_dir}/{station_id}.nc')


def test_formats():
    config = base_config()
    config.start_date = '2015-07-28'
    config.end_date = '2020-12-31'
    config.timestep = 'h'

    meta = pd.read_csv(f'{pytest.DATA_DIR}/meteo/rofental/csv/stations.csv', index_col=0)
    meta = meta.sort_index()

    meteo = {}
    for fmt in ('netcdf', 'csv'):
        config.input_data.meteo.format = fmt
        config.input_data.meteo.dir = f'{pytest.DATA_DIR}/meteo/rofental/{fmt}'
        model = oa.OpenAmundsen(config)
        model.initialize()
        ds = model.meteo.sortby('station')
        assert ds.time.to_index().equals(model.dates)
        assert_array_equal(meta['name'], ds.station_name)
        assert_array_equal(meta['alt'], ds.alt)
        meteo[fmt] = ds

    xr.testing.assert_allclose(meteo['netcdf'], meteo['csv'])


def test_format_memory(tmp_path):
    config = base_config()
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo1 = model.meteo

    config.input_data.meteo.format = 'memory'
    model = oa.OpenAmundsen(config)
    model.initialize(meteo=oa.forcing.strip_point_dataset(meteo1))
    meteo2 = model.meteo
    assert meteo1.identical(meteo2)

    config.start_date = meteo1.indexes['time'][1]
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize(meteo=oa.forcing.strip_point_dataset(meteo1))


def test_format_callback():
    def callback(model):
        model.state.meteo.temp[:] = meteo['temp'][model.date_idx]
        model.state.meteo.precip[:] = meteo['precip'][model.date_idx]
        model.state.meteo.wind_speed[:] = meteo['wind_speed'][model.date_idx]
        model.state.meteo.rel_hum[:] = meteo['rel_hum'][model.date_idx]
        model.state.meteo.cloud_factor[:] = meteo['cloud_factor'][model.date_idx]
        model.state.meteo.cloud_fraction[:] = meteo['cloud_fraction'][model.date_idx]

    config = base_config()
    config.start_date = '2019-12-01'
    config.end_date = '2019-12-15'

    model = oa.OpenAmundsen(config)
    model.initialize()

    for category in model.state.categories:
        for var_name in model.state[category]._meta.keys():
            full_var_name = f'{category}.{var_name}'
            config.output_data.grids.variables.append({
                'var': full_var_name,
                'name': full_var_name,
            })

    meteo = {
        'temp': [],
        'precip': [],
        'wind_speed': [],
        'rel_hum': [],
        'cloud_factor': [],
        'cloud_fraction': [],
    }

    model = oa.OpenAmundsen(config)
    model.initialize()
    for date in model.dates:
        model.run_single()
        meteo['temp'].append(model.state.meteo.temp.copy())
        meteo['precip'].append(model.state.meteo.precip.copy())
        meteo['wind_speed'].append(model.state.meteo.wind_speed.copy())
        meteo['rel_hum'].append(model.state.meteo.rel_hum.copy())
        meteo['cloud_factor'].append(model.state.meteo.cloud_factor.copy())
        meteo['cloud_fraction'].append(model.state.meteo.cloud_fraction.copy())
    data_ref = model.gridded_output.data.copy()

    config.input_data.meteo = {'format': 'callback'}
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()
    model.initialize(meteo_callback=callback)
    model.run()
    data_cb = model.gridded_output.data.copy()

    assert data_ref.identical(data_cb)


def test_no_files_found(tmp_path):
    config = base_config()
    config.input_data.meteo.dir = str(tmp_path)

    for fmt in ('netcdf', 'csv'):
        config.input_data.meteo.format = fmt
        model = oa.OpenAmundsen(config)
        with pytest.raises(errors.MeteoDataError):
            model.initialize()


def test_missing_csv_metadata_columns(tmp_path):
    config = base_config()
    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    meta[['name', 'x', 'y']].loc[['bellavista']].to_csv(tmp_path / 'stations.csv')
    (tmp_path / 'bellavista.csv').symlink_to(p_orig / 'bellavista.csv')

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_records_inbetween(tmp_path):
    config = base_config()
    config.end_date = '2020-04-30'
    config.timestep = 'h'
    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')

    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    meta.loc[['bellavista']].to_csv(tmp_path / 'stations.csv')

    df = pd.read_csv(p_orig / 'bellavista.csv', index_col=0, parse_dates=True)
    df = df.loc[config.start_date:config.end_date]
    df = df.drop(df.index[37:38])
    df.to_csv(tmp_path / 'bellavista.csv')

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_records_start_end():
    config = base_config()
    config.start_date = '1900-01-01'
    config.end_date = '2100-12-31'

    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_station_selection(tmp_path):
    config = base_config()
    model = oa.OpenAmundsen(config)
    model.initialize()

    dummy_station = model.meteo.isel(station=0).copy()
    dummy_station['station'] = 'dummy'
    dummy_station['station_name'] = 'Dummy station'
    dummy_station['lon'] = 11.7
    dummy_station['lat'] = 47.21
    ds = xr.concat([model.meteo, dummy_station], dim='station')
    meteo_to_netcdf(ds, tmp_path)

    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.bounds = 'grid'
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista', 'latschbloder', 'proviantdepot'],
    )

    config.input_data.meteo.bounds = 'global'
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista', 'dummy', 'latschbloder', 'proviantdepot'],
    )

    config.input_data.meteo.bounds = [
        636800,
        5182550,
        636850,
        5182580,
    ]
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista'],
    )

    config.input_data.meteo.bounds = [
        636800,
        5182550,
        636801,
        5182551,
    ]
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.bounds = 'global'
    config.input_data.meteo.exclude = ['bellavista', 'latschbloder']
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['dummy', 'proviantdepot'],
    )

    config.input_data.meteo.bounds = 'grid'
    config.input_data.meteo.exclude = ['bellavista', 'latschbloder', 'proviantdepot']
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.input_data.meteo.bounds = 'grid'
    config.input_data.meteo.exclude = []
    config.input_data.meteo.include = ['dummy']
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert_array_equal(
        sorted(model.meteo.station.values),
        ['bellavista', 'dummy', 'latschbloder', 'proviantdepot'],
    )

    config.input_data.meteo.include = []
    config.input_data.meteo.exclude = ['dummy', 'dummy2']
    model = oa.OpenAmundsen(config)
    model.initialize()

    config.input_data.meteo.include = ['dummy', 'dummy2']
    config.input_data.meteo.exclude = []
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_missing_variables(tmp_path):
    ds = xr.load_dataset(f'{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc')
    ds = ds.drop_vars(['tas', 'hurs', 'rsds', 'wss', 'wind_dir'])
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')

    config = base_config()
    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()

    assert 'temp' in model.meteo
    assert 'precip' in model.meteo
    assert 'rel_hum' in model.meteo
    assert 'sw_in' in model.meteo
    assert 'wind_speed' in model.meteo
    assert 'wind_dir' not in model.meteo


def test_netcdf_precip_units(tmp_path):
    config = base_config()
    config.input_data.meteo.dir = str(tmp_path)

    ds = xr.load_dataset(f'{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc')
    ds.pr.attrs['units'] = 'kg m-2 s-1'
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo1 = model.meteo.copy()

    ds['pr'] *= 3600
    ds.pr.attrs['units'] = 'kg m-2'
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')
    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo2 = model.meteo.copy()

    xr.testing.assert_allclose(meteo1.precip, meteo2.precip)

    ds.pr.attrs['units'] = 'mm h-1'
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()


def test_crs(tmp_path):
    config = base_config()

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds1 = model.meteo[['lon', 'lat', 'x', 'y']].sortby('station')

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    x_new, y_new = oa.util.transform_coords(meta.x, meta.y, 'epsg:32632', 'epsg:3416')
    meta.x = x_new
    meta.y = y_new
    meta.to_csv(tmp_path / 'stations.csv')

    for station_id in meta.index:
        (tmp_path / f'{station_id}.csv').symlink_to(p_orig / f'{station_id}.csv')

    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.crs = 'epsg:3416'

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds2 = model.meteo[['lon', 'lat', 'x', 'y']].sortby('station')

    xr.testing.assert_allclose(ds1, ds2)


def test_grid_cell_assignment(tmp_path):
    config = base_config()
    config.start_date = '2020-01-01'
    config.end_date = '2020-01-01'

    model = oa.OpenAmundsen(config)
    model.initialize()
    grid_xs = model.grid.xs
    grid_ys = model.grid.ys

    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)

    station_coords_colrows = [  # (x, y, col, row)
        (grid_xs[0], grid_ys[0], 0, 0),
        (grid_xs[0] + 1, grid_ys[0] - 1, 0, 0),
        (grid_xs[1] + 1, grid_ys[0] - 1, 1, 0),
        (grid_xs[1] + 1, grid_ys[3] - 1, 1, 3),
        (grid_xs[-1] + 1, grid_ys[-1] - 1, len(grid_xs) - 1, len(grid_ys) - 1),
        (grid_xs[0], grid_ys[-1], 0, len(grid_ys) - 1),
    ]

    meta = pd.DataFrame(
        index=[f'dummy{i}' for i in range(len(station_coords_colrows))],
        data={
            'name': None,
            'x': [s[0] for s in station_coords_colrows],
            'y': [s[1] for s in station_coords_colrows],
            'alt': 0.,
        },
    )
    meta.to_csv(tmp_path / 'stations.csv')

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    for station_id in meta.index:
        (tmp_path / f'{station_id}.csv').symlink_to(p_orig / 'proviantdepot.csv')

    model = oa.OpenAmundsen(config)
    model.initialize()
    ds = model.meteo

    for i, s in enumerate(station_coords_colrows):
        ds_s = ds.sel(station=f'dummy{i}')
        expected_col = s[2]
        expected_row = s[3]
        assert int(ds_s.col) == expected_col
        assert int(ds_s.row) == expected_row


@pytest.mark.parametrize('fmt', ['netcdf', 'csv'])
def test_slice_and_resample(fmt, tmp_path):
    def meteo_to_df(model):
        return (
            model.meteo
            .sel(station='bellavista')
            .to_dataframe()
            [params]
        ).astype(float)

    def matches_start_and_end_date(model):
        idx = model.meteo.indexes['time']
        return (
            idx[0] == model.config.start_date
            and idx[-1] == model.config.end_date
        )

    params = ['temp', 'precip', 'rel_hum', 'sw_in', 'wind_speed']

    agg_funcs_inst = {p: lambda s: s.iloc[-1] for p in params}
    agg_funcs_inst['precip'] = pd.Series.sum
    agg_funcs_res = {p: pd.Series.mean for p in params}
    agg_funcs_res['precip'] = pd.Series.sum

    config = base_config()
    config.start_date = '2015-11-20'
    config.end_date = '2015-11-30'
    config.timestep = 'h'
    config.input_data.meteo.format = fmt
    config.input_data.meteo.dir = f'{pytest.DATA_DIR}/meteo/rofental/{fmt}'
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_h = meteo_to_df(model)

    config.timestep = '3h'
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 13:00':'2015-11-30 15:00'].agg(agg_funcs_inst, skipna=False),
        df_res.loc['2015-11-30 15:00'],
        check_exact=False,
        check_names=False,
    )
    assert np.isnan(df_res.loc['2015-11-26 15:00'].precip)
    assert_allclose(
        df_h.loc[:df_res.index[-1]].precip.sum(),
        df_res.precip.sum(),
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 13:00':'2015-11-30 15:00'].agg(agg_funcs_res, skipna=False),
        df_res.loc['2015-11-30 15:00'],
        check_exact=False,
        check_names=False,
    )
    assert np.isnan(df_res.loc['2015-11-26 15:00'].precip)

    config.start_date = '2015-11-25'
    config.end_date = '2015-11-29'
    config.timestep = 'D'
    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-25 01:00':'2015-11-26 00:00'].agg(agg_funcs_res, skipna=False),
        df_res.loc['2015-11-25'],
        check_exact=False,
        check_names=False,
    )
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-29 01:00':'2015-11-30 00:00'].agg(agg_funcs_res, skipna=False),
        df_res.loc['2015-11-29'],
        check_exact=False,
        check_names=False,
    )

    config.start_date = '2015-11-01'
    config.end_date = '2015-11-09'
    config.timestep = '2D'
    wf = config.output_data.timeseries.write_freq
    config.output_data.timeseries.write_freq = '10D'
    model = oa.OpenAmundsen(config)
    config.output_data.timeseries.write_freq = wf
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.start_date = '2015-11-01 01:00'
    config.end_date = '2015-11-30 22:00'
    config.timestep = '3h'
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 14:00':'2015-11-30 16:00'].agg(agg_funcs_inst, skipna=False),
        df_res.loc['2015-11-30 16:00'],
        check_exact=False,
        check_names=False,
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    assert matches_start_and_end_date(model)
    df_res = meteo_to_df(model)
    pd.testing.assert_series_equal(
        df_h.loc['2015-11-30 14:00':'2015-11-30 16:00'].agg(agg_funcs_res, skipna=False),
        df_res.loc['2015-11-30 16:00'],
        check_exact=False,
        check_names=False,
    )

    config.start_date = '2015-11-01 01:37'
    config.end_date = '2015-11-30 22:37'
    config.timestep = '3h'
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    config.start_date = '2015-11-29'
    config.end_date = '2015-11-30'
    config.timestep = '10min'
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.MeteoDataError):
        model.initialize()

    # Wind direction
    config.start_date = '2015-12-20'
    config.end_date = '2015-12-30'
    config.timestep = 'h'
    config.input_data.meteo.aggregate_when_downsampling = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_h = model.meteo.sel(station='bellavista').to_dataframe()[['wind_speed', 'wind_dir']]

    config.timestep = '3h'
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = model.meteo.sel(station='bellavista').to_dataframe()[['wind_speed', 'wind_dir']]
    assert np.allclose(
        df_h['wind_dir'].loc['2015-12-20 15:00'],
        df_res['wind_dir'].loc['2015-12-20 15:00'],
    )

    config.input_data.meteo.aggregate_when_downsampling = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    df_res = model.meteo.sel(station='bellavista').to_dataframe()[['wind_speed', 'wind_dir']]
    wind_us, wind_vs = oameteo.wind_to_uv(
        df_h['wind_speed'].loc['2015-12-20 13:00':'2015-12-20 15:00'],
        df_h['wind_dir'].loc['2015-12-20 13:00':'2015-12-20 15:00'],
    )
    wind_u_mean = wind_us.mean()
    wind_v_mean = wind_vs.mean()
    _, wind_dir = oameteo.wind_from_uv(wind_u_mean, wind_v_mean)
    assert np.allclose(
        df_res['wind_dir'].loc['2015-12-20 15:00'],
        wind_dir,
    )


def test_aggregation_when_using_csv_input_and_start_date_before_station_start_date(tmp_path):
    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/csv')
    meta = pd.read_csv(p_orig / 'stations.csv', index_col=0)
    meta.loc[['bellavista', 'latschbloder']].to_csv(tmp_path / 'stations.csv')
    (tmp_path / 'latschbloder.csv').symlink_to(p_orig / 'latschbloder.csv')
    df = pd.read_csv(p_orig / 'bellavista.csv', index_col=0, parse_dates=True)
    df.loc['2015-12-22':].to_csv(tmp_path / 'bellavista.csv')

    config = base_config()
    config.start_date = '2015-12-20'
    config.end_date = '2015-12-30'
    config.timestep = 'D'
    config.input_data.meteo.format = 'csv'
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.aggregate_when_downsampling = True

    model = oa.OpenAmundsen(config)
    model.initialize()


def test_non_hourly_input(tmp_path):
    config = base_config()
    config.start_date = '2015-11-20'
    config.end_date = '2015-11-30'
    config.timestep = '3h'

    model = oa.OpenAmundsen(config)
    model.initialize()
    meteo_to_netcdf(model.meteo, tmp_path)
    ds1 = model.meteo

    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds2 = model.meteo

    xr.testing.assert_allclose(ds1, ds2)


@pytest.mark.parametrize('aggregate', [False, True])
def test_resample_with_non_matching_start_date(aggregate, tmp_path):
    ds = xr.load_dataset(f'{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc')
    ds = ds.sel(time=slice('2020-11-03 02:00', None))
    ds.to_netcdf(tmp_path / 'proviantdepot.nc')

    p_orig = Path(f'{pytest.DATA_DIR}/meteo/rofental/netcdf')
    for station_id in ('bellavista', 'latschbloder'):
        (tmp_path / f'{station_id}.nc').symlink_to(p_orig / f'{station_id}.nc')

    config = base_config()
    config.start_date = '2020-11-01'
    config.end_date = '2020-11-30'
    config.timestep = '3h'
    config.input_data.meteo.dir = str(tmp_path)
    config.input_data.meteo.aggregate_when_downsampling = aggregate

    model = oa.OpenAmundsen(config)
    model.initialize()
    assert model.meteo.indexes['time'].equals(model.dates)


def test_filters(tmp_path):
    config = base_config()
    config.start_date = '2015-07-28'
    config.end_date = '2020-12-31'
    config.timestep = 'h'
    default_filters = config.input_data.meteo.filters

    config.input_data.meteo.filters = []
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds_unfiltered = model.meteo

    config.input_data.meteo.filters = default_filters
    model = oa.OpenAmundsen(config)
    model.initialize()
    xr.testing.assert_identical(ds_unfiltered, model.meteo)

    ds_unfiltered['temp'].loc['bellavista', '2015-08-01 12:00':'2015-08-02 23:00'] = 15
    ds_unfiltered['rel_hum'].loc['bellavista', '2015-08-03 12:00'] = 100
    ds_unfiltered['rel_hum'].loc['bellavista', '2015-08-03 13:00'] = 100.1
    ds_unfiltered['precip'].loc['bellavista', '2015-08-03 12:00'] = -10
    ds_unfiltered['precip'].loc['bellavista', '2015-08-03 13:00'] = 10000

    meteo_to_netcdf(ds_unfiltered, tmp_path)
    config.input_data.meteo.dir = str(tmp_path)
    model = oa.OpenAmundsen(config)
    model.initialize()
    ds_filtered = model.meteo
    assert all(np.isnan(ds_filtered['temp'].loc['bellavista', '2015-08-01 12:00':'2015-08-02 23:00']))
    assert ds_filtered['rel_hum'].loc['bellavista', '2015-08-03 12:00'] == 100
    assert np.isnan(ds_filtered['rel_hum'].loc['bellavista', '2015-08-03 13:00'])
    assert np.isnan(ds_filtered['precip'].loc['bellavista', '2015-08-03 12:00'])
    assert ds_filtered['precip'].loc['bellavista', '2015-08-03 13:00'] == 10000
    xr.testing.assert_identical(
        ds_unfiltered.sel(time=slice('2015-08-04 00:00', None)),
        ds_filtered.sel(time=slice('2015-08-04 00:00', None)),
    )



DAYS_PER_YEAR = 365.25
HOURS_PER_DAY = 24
MINUTES_PER_HOUR = 60
SECONDS_PER_HOUR = 3600

STANDARD_TIMEZONE_WIDTH = 15
SUN_DEGREES_PER_HOUR = 15  # 360/24
MINUTES_PER_DEGREE_OF_EARTH_ROTATION = 4  # earth rotates 1 degree every 4 minutes
SOLAR_CONSTANT = 1366.1  # W m-2

STANDARD_ATMOSPHERE = 101_325  # Pa
ATMOSPHERIC_LAPSE_RATE = 6.5e-3  # K m-1
GRAVITATIONAL_ACCELERATION = 9.80665  # m s-2

T0 = 273.15  # K

MOLAR_MASS_DRY_AIR = 0.02896968  # kg mol-1
MOLAR_MASS_WATER = 0.0180153  # kg mol-1
UNIVERSAL_GAS_CONSTANT = 8.314462618  # J mol-1 K-1
GAS_CONSTANT_DRY_AIR = 287.058  # J kg-1 K-1
SPEC_GAS_CONSTANT_WATER_VAPOR = 461.52  # J kg-1 K-1
STANDARD_SEA_LEVEL_TEMPERATURE = 15 + T0  # K

SPEC_HEAT_CAP_DRY_AIR = 1004.68506  # specific heat capacity of dry air (J kg-1 K-1)
SPEC_HEAT_CAP_ICE = 2100.  # specific heat capacity of ice (J kg-1 K-1)
SPEC_HEAT_CAP_WATER = 4180.  # specific heat capacity of water (J kg-1 K-1)

LATENT_HEAT_OF_VAPORIZATION = 2.501e6  # J kg-1 (= latent heat of condensation)
LATENT_HEAT_OF_FUSION = 0.334e6  # J kg-1
LATENT_HEAT_OF_SUBLIMATION = LATENT_HEAT_OF_VAPORIZATION + LATENT_HEAT_OF_FUSION  # J kg-1

STEFAN_BOLTZMANN = 5.670374419e-8  # Stefan-Boltzmann constant (W m-2 K-4)
VON_KARMAN = 0.40  # von Karman constant

WATER_DENSITY = 1000.  # density of water (kg m-3)
ICE_DENSITY = 917.  # density of ice (kg m-3)

# Soil constants
VOL_HEAT_CAP_SAND = 2.128e6  # volumetric heat capacity of sand (J m-3 K-1)
VOL_HEAT_CAP_CLAY = 2.385e6  # volumetric heat capacity of clay (J m-3 K-1)

# Thermal conductivities
THERM_COND_AIR = 0.025  # thermal conductivity of air (W m-1 K-1)
THERM_COND_CLAY = 1.16  # thermal conductivity of clay (W m-1 K-1)
THERM_COND_ICE = 2.24  # thermal conducivity of ice (W m-1 K-1)
THERM_COND_SAND = 1.57  # thermal conductivity of sand (W m-1 K-1)
THERM_COND_WATER = 0.56  # thermal conductivity of water (W m-1 K-1)

CRS_WGS84 = 'epsg:4326'

METEO_VAR_METADATA = {
    'temp': {
        'standard_name': 'air_temperature',
        'units': 'K',
    },
    'precip': {
        'standard_name': 'precipitation_amount',
        'units': 'kg m-2',
    },
    'rel_hum': {
        'standard_name': 'relative_humidity',
        'units': '%',
    },
    'sw_in': {
        'standard_name': 'surface_downwelling_shortwave_flux_in_air',
        'units': 'W m-2',
    },
    'wind_speed': {
        'standard_name': 'wind_speed',
        'units': 'm s-1',
    },
    'wind_speed_gust': {
        'standard_name': 'wind_speed_of_gust',
        'units': 'm s-1',
    },
    'wind_dir': {
        'standard_name': 'wind_from_direction',
        'units': 'degree',
    },
    'cloud_fraction': {
        'standard_name': 'cloud_area_fraction',
        'units': '%',
    },
}
MINIMUM_REQUIRED_METEO_VARS = ['temp', 'precip', 'rel_hum', 'sw_in', 'wind_speed']

NETCDF_VAR_MAPPINGS = {
    'tas': 'temp',
    'pr': 'precip',
    'hurs': 'rel_hum',
    'rsds': 'sw_in',
    'wss': 'wind_speed',
    'wind_dir': 'wind_dir',
    'wind_speed_gust': 'wind_speed_gust',
}

# Mappings of internal variable names to interpolation config keys (e.g.
# config['meteo']['interpolation']['temperature'])
INTERPOLATION_CONFIG_PARAM_MAPPINGS = {
    'temp': 'temperature',
    'precip': 'precipitation',
    'rel_hum': 'humidity',
}

ALLOWED_METEO_VAR_RANGES = {
    'temp': (0, None),
    'precip': (0, None),
    'rel_hum': (0, 100),
    'sw_in': (0, None),
    'wind_speed': (0.1, None),
    'wind_speed_gust': (0.1, None),
    'cloud_fraction': (0, 100),
}



import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import openamundsen as oa
import pandas as pd


def test_slope():
    slope, _ = oa.terrain.slope_aspect(np.zeros((3, 3)), 1)
    assert_almost_equal(slope[1][1], 0)

    dem = np.array([
        [50, 45, 50],
        [30, 30, 30],
        [8, 10, 10],
    ], dtype=float)  # from https://desktop.arcgis.com/en/arcmap/10.3/tools/spatial-analyst-toolbox/how-slope-works.htm
    slope, _ = oa.terrain.slope_aspect(dem, 5)
    assert_almost_equal(slope[1][1], 75.25765769)


def test_aspect():
    dem = np.array([
        [0, 0, 0],
        [1, 1, 1],
        [2, 2, 2],
    ], dtype=float)

    _, aspect = oa.terrain.slope_aspect(dem, 1)
    assert aspect[1][1] == 0  # north facing

    _, aspect = oa.terrain.slope_aspect(np.rot90(dem), 1)
    assert aspect[1][1] == 270  # west facing

    _, aspect = oa.terrain.slope_aspect(np.rot90(dem, k=2), 1)
    assert aspect[1][1] == 180  # south facing

    _, aspect = oa.terrain.slope_aspect(np.rot90(dem, k=3), 1)
    assert aspect[1][1] == 90  # east facing

    dem = np.array([
        [101, 92, 85],
        [101, 92, 85],
        [101, 91, 84],
    ], dtype=float)  # from https://desktop.arcgis.com/en/arcmap/10.3/tools/3d-analyst-toolbox/how-aspect-works.htm
    _, aspect = oa.terrain.slope_aspect(dem, 1)
    assert_almost_equal(aspect[1][1], 92.64254529)



from numba import njit, prange
import numpy as np
from openamundsen import constants, meteo
from openamundsen.modules.snow import CryoLayerID


def surface_properties(model):
    """
    Update surface properties following [1].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015

    .. [2] Essery, R. L. H., Best, M. J., & Cox, P. M. (2001). MOSES 2.2
       Technical Documentation, Tech. rep., Hadley Centre, Met Office.
       http://jules.jchmr.org/sites/default/files/HCTN_30.pdf
    """
    s = model.state

    s.surface.albedo[:] = np.where(
        s.snow.area_fraction > 0,
        (
            s.snow.area_fraction * s.snow.albedo
            + (1 - s.snow.area_fraction) * model.config.soil.albedo
        ),
        model.config.soil.albedo,
    )

    s.surface.roughness_length[:] = (
        model.config.snow.roughness_length**s.snow.area_fraction
        * model.config.soil.roughness_length**(1 - s.snow.area_fraction)
    )
    calc_turbulent_exchange_coefficient(model)

    if model.config.snow.model == 'multilayer':
        # Calculate surface conductance (eq. (35) from [2])
        # (the constant 1/100 therein corresponds to
        # model.config.soil.saturated_soil_surface_conductance; the clipping of (theta_1/theta_c)**2
        # to 1 is taken from FSM)
        s.surface.conductance[:] = model.config.soil.saturated_soil_surface_conductance * (
            np.maximum(
                (
                    s.soil.frac_unfrozen_moisture_content[0, :]
                    * s.soil.vol_moisture_content_sat / s.soil.vol_moisture_content_crit
                )**2,
                1,
            )
        )

        surface_layer_properties(model)


def surface_layer_properties(model):
    """
    Update surface layer properties following [1].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Cox, P. M., Betts, R. A., Bunton, C. B., Essery, R. L. H., Rowntree,
       P. R., & Smith, J. (1999). The impact of new land surface physics on the
       GCM simulation of climate and climate sensitivity. Climate Dynamics, 15(3),
       183–203. https://doi.org/10.1007/s003820050276
    """
    s = model.state

    s.surface.thickness[:] = np.maximum(s.snow.thickness[0, :], s.soil.thickness[0, :])
    s.surface.layer_temp[:] = (
        s.soil.temp[0, :]
        + (s.snow.temp[0, :] - s.soil.temp[0, :])
        * s.snow.thickness[0, :] / s.soil.thickness[0, :]
    )

    pos1 = s.snow.thickness[0, :] <= s.soil.thickness[0, :] / 2
    pos2 = s.snow.thickness[0, :] > s.soil.thickness[0, :]

    # Effective surface thermal conductivity (eq. (79))
    s.surface.therm_cond[:] = s.snow.therm_cond[0, :]
    s.surface.therm_cond[pos1] = (
        s.soil.thickness[0, pos1]
        / (
            2 * s.snow.thickness[0, pos1] / s.snow.therm_cond[0, pos1]
            + (
                s.soil.thickness[0, pos1] - 2 * s.snow.thickness[0, pos1]
            ) / s.soil.therm_cond[0, pos1]
        )
    )

    s.surface.layer_temp[pos2] = s.snow.temp[0, pos2]


def energy_balance(model):
    snow_model = model.config.snow.model

    if snow_model == 'multilayer':
        multilayer_energy_balance(model)
    elif snow_model == 'cryolayers':
        cryo_layer_energy_balance(model)


def multilayer_energy_balance(model):
    """
    Calculate the surface energy balance for the multilayer model
    following [1].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015

    .. [2] Essery, R. L. H., Best, M. J., & Cox, P. M. (2001). MOSES 2.2
       Technical Documentation, Tech. rep., Hadley Centre, Met Office.
       http://jules.jchmr.org/sites/default/files/HCTN_30.pdf
    """
    s = model.state
    roi = model.grid.roi

    calc_saturation_specific_humidity(model, roi)
    calc_moisture_availability(model, roi)
    calc_latent_heat(model, roi)

    # Calculate surface energy balance without melt
    s.snow.melt[roi] = 0
    calc_radiation_balance(model, roi)
    calc_surface_flux(model, roi)
    calc_turbulent_fluxes(model, roi)
    calc_advective_heat(model, roi)

    (
        surf_temp_change,
        surf_moisture_flux_change,
        surf_heat_flux_change,
        sens_heat_flux_change,
    ) = solve_energy_balance(model, roi)

    # Calculate melt
    melties_roi = (
        ((s.surface.temp[roi] + surf_temp_change) > constants.T0)
        & (s.snow.ice_content[0, roi] > 0)
    )
    if melties_roi.any():
        melties = model.roi_mask_to_global(melties_roi)
        s.snow.melt[melties] = s.snow.ice_content[:, melties].sum(axis=0)

        (
            surf_temp_change[melties_roi],
            surf_moisture_flux_change[melties_roi],
            surf_heat_flux_change[melties_roi],
            sens_heat_flux_change[melties_roi],
        ) = solve_energy_balance(model, melties)

        partial_melties_roi = melties_roi & (s.surface.temp[roi] + surf_temp_change < constants.T0)
        if partial_melties_roi.any():
            partial_melties = model.roi_mask_to_global(partial_melties_roi)

            s.surface.temp[partial_melties] = constants.T0

            calc_latent_heat(model, partial_melties)
            calc_saturation_specific_humidity(model, partial_melties)
            calc_radiation_balance(model, partial_melties)  # update net radiation
            calc_surface_flux(model, partial_melties)
            calc_turbulent_fluxes(model, partial_melties)
            calc_advective_heat(model, partial_melties)

            s.snow.melt[partial_melties] = (
                (
                    s.meteo.net_radiation[partial_melties]
                    + s.surface.sens_heat_flux[partial_melties]
                    + s.surface.lat_heat_flux[partial_melties]
                    + s.surface.advective_heat_flux[partial_melties]
                    - s.surface.heat_flux[partial_melties]
                ) / constants.LATENT_HEAT_OF_FUSION * model.timestep
            ).clip(min=0)

            surf_temp_change[partial_melties_roi] = 0.
            surf_moisture_flux_change[partial_melties_roi] = 0.
            surf_heat_flux_change[partial_melties_roi] = 0.
            sens_heat_flux_change[partial_melties_roi] = 0.

    # Update surface temperature and fluxes
    s.surface.temp[roi] += surf_temp_change
    s.surface.moisture_flux[roi] += surf_moisture_flux_change
    s.surface.heat_flux[roi] += surf_heat_flux_change
    s.surface.sens_heat_flux[roi] += sens_heat_flux_change
    calc_radiation_balance(model, roi)
    calc_latent_heat(model, roi)
    calc_turbulent_fluxes(model, roi, sensible=False)


def cryo_layer_energy_balance(model):
    """
    Calculate the surface energy balance for the cryo layer model following [1-2].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Strasser, U. (2008). Die Modellierung der Gebirgsschneedecke im
       Nationalpark Berchtesgaden. Modelling of the mountain snow cover in the
       Berchtesgaden National Park, Berchtesgaden National Park research report,
       No. 55, Berchtesgaden.

    .. [2] Hanzer, F., Helfricht, K., Marke, T., & Strasser, U. (2016).
       Multilevel spatiotemporal validation of snow/ice mass balance and runoff
       modeling in glacierized catchments. The Cryosphere, 10(4), 1859–1881.
       https://doi.org/10.5194/tc-10-1859-2016
    """
    s = model.state
    roi = model.grid.roi

    s.surface.temp[roi] = s.meteo.temp[roi]

    snowies_roi = s.snow.swe[roi] > 0.
    melties_roi = snowies_roi & (s.meteo.temp[roi] >= constants.T0)
    frosties_roi = snowies_roi & (s.meteo.temp[roi] < constants.T0)
    snowies = model.roi_mask_to_global(snowies_roi)
    melties = model.roi_mask_to_global(melties_roi)
    frosties = model.roi_mask_to_global(frosties_roi)
    snow_freeies = model.roi_mask_to_global(~snowies_roi)

    calc_saturation_specific_humidity(model, roi)
    calc_moisture_availability(model, roi)
    calc_latent_heat(model, roi)

    s.snow.melt[roi] = 0.
    s.snow.refreezing[roi] = 0.

    s.surface.heat_flux[snowies] = model.config.snow.cryolayers.surface_heat_flux
    s.surface.heat_flux[snow_freeies] = np.nan

    # Where air temperature >= 0 °C -> potential melt, no iteration
    s.surface.temp[melties] = constants.T0
    available_melt_time = np.zeros(roi.shape)
    en_bal = np.full(roi.shape, np.nan)
    available_melt_time[melties] = model.timestep  # contains the time (in seconds) available for melt in this time step for each pixel

    for layer_num in range(model.snow.num_cryo_layers):
        possible_melties = model.roi_mask_to_global(
            melties_roi
            & (s.snow.thickness[layer_num, roi] > 0)
            & (available_melt_time[roi] > 0)
        )

        en_bal[possible_melties] = energy_balance_remainder(
            model,
            possible_melties,
            s.surface.temp[possible_melties],
        )

        layer_melties = model.roi_mask_to_global(possible_melties[roi] & (en_bal[roi] > 0.))

        layer_we = (
            s.snow.ice_content[layer_num, :]
            # + s.snow.liquid_water_content[layer_num, :]
            + s.snow.cold_content[layer_num, :]
        )

        layer_melt_time = np.zeros(roi.shape)
        layer_melt_time[layer_melties] = (  # time needed to melt the entire layer
            constants.LATENT_HEAT_OF_FUSION
            * layer_we[layer_melties]
            / en_bal[layer_melties]
        )
        total_melties = model.roi_mask_to_global(
            layer_melties[roi]
            & (layer_melt_time[roi] < available_melt_time[roi])
        )
        partial_melties = model.roi_mask_to_global(
            layer_melties[roi]
            & (layer_melt_time[roi] >= available_melt_time[roi])
        )

        # Process pixels where the entire layer melts (and energy is left to melt the lower layers)
        s.snow.melt[total_melties] += s.snow.ice_content[layer_num, total_melties]
        s.snow.cold_content[layer_num, total_melties] = 0.

        # Process pixels where the available energy cannot melt the entire layer
        layer_melt = (  # actual melt and cold content reduction
            en_bal[partial_melties]
            * available_melt_time[partial_melties]
            / constants.LATENT_HEAT_OF_FUSION
        )
        s.snow.cold_content[layer_num, partial_melties] -= layer_melt
        actual_layer_melt = -1 * np.minimum(s.snow.cold_content[layer_num, partial_melties], 0.)  # actual melt not used for reducing the cold content
        s.snow.melt[partial_melties] += actual_layer_melt
        s.snow.cold_content[layer_num, partial_melties] = (
            s.snow.cold_content[layer_num, partial_melties].clip(min=0)
        )

        available_melt_time[melties] -= layer_melt_time[melties]

    # Refreeze liquid water and build up cold content
    en_bal[frosties] = energy_balance_remainder(model, frosties, constants.T0)
    refreezing_factor = model.config.snow.cryolayers.refreezing_factor
    available_cc = (
        -1
        * en_bal * model.timestep / constants.LATENT_HEAT_OF_FUSION
        * refreezing_factor
    ).clip(min=0)  # kg m-2
    cold_holding_capacity = model.config.snow.cryolayers.cold_holding_capacity
    for layer_num in range(model.snow.num_cryo_layers):
        # No cold content for firn and ice
        if layer_num not in (CryoLayerID.NEW_SNOW, CryoLayerID.OLD_SNOW):
            continue

        cold_conties = model.global_mask(
            frosties_roi
            & (s.snow.thickness[layer_num, roi] > 0)
            & (available_cc[roi] > 0)
        )

        refreeze_amount = np.minimum(
            s.snow.liquid_water_content[layer_num, cold_conties],
            available_cc[cold_conties],
        )
        s.snow.liquid_water_content[layer_num, cold_conties] -= refreeze_amount
        s.snow.ice_content[layer_num, cold_conties] += refreeze_amount
        s.snow.refreezing[cold_conties] += refreeze_amount
        available_cc[cold_conties] -= refreeze_amount

        max_layer_cc = cold_holding_capacity * (
            s.snow.ice_content[layer_num, cold_conties]
            + s.snow.liquid_water_content[layer_num, cold_conties]
        )
        cc_buildup = np.minimum(
            available_cc[cold_conties],
            max_layer_cc - s.snow.cold_content[layer_num, cold_conties],
        )
        s.snow.cold_content[layer_num, cold_conties] += cc_buildup
        available_cc[cold_conties] -= cc_buildup

    # Iteration for calculating snow surface temperature
    iterate_surface_temperature(model, frosties)

    calc_radiation_balance(model, roi)
    calc_turbulent_fluxes(model, roi)
    calc_advective_heat(model, roi)


def stability_factor(
    temp,
    surface_temp,
    wind_speed,
    snow_cover_fraction,
    snow_roughness_length,
    snow_free_roughness_length,
    temp_measurement_height,
    wind_measurement_height,
    stability_adjustment_parameter,
):
    """
    Calculate the atmospheric stability factor from [1] (eq. (25)).

    Parameters
    ----------
    temp : ndarray
        Air temperature (K).

    surface_temp : ndarray
        Surface temperature (K).

    wind_speed : ndarray
        Wind speed (m s-1).

    snow_cover_fraction : ndarray
        Snow cover fraction.

    snow_roughness_length : float
        Roughness length of snow-covered ground (m).

    snow_free_roughness_length : float
        Roughness length of snow-free ground (m).

    temp_measurement_height : float
        Temperature measurement height (m).

    wind_measurement_height : float
        Wind measurement height (m).

    stability_adjustment_parameter : float
        Atmospheric stability adjustment parameter (b_h from [1]).

    Returns
    -------
    stability_factor : ndarray
        Atmospheric stability factor.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    # Replace zero wind speeds (would lead to a divide by zero) with a plausible value
    wind_speed = wind_speed.copy()
    wind_speed[wind_speed == 0.] = 0.5

    # Bulk Richardson number (eq. (24))
    richardson = (
        constants.GRAVITATIONAL_ACCELERATION
        * wind_measurement_height**2
        * (temp - surface_temp)
        / (temp_measurement_height * temp * wind_speed**2)
    )

    # Surface momentum roughness length (eq. (23))
    momentum_roughness_length = (
        snow_roughness_length**snow_cover_fraction
        * snow_free_roughness_length**(1 - snow_cover_fraction)
    )

    pos = richardson >= 0

    # eq. (26)
    c = (
        3 * stability_adjustment_parameter**2 * constants.VON_KARMAN**2
        * np.sqrt(wind_measurement_height / momentum_roughness_length[~pos])
        / (np.log(wind_measurement_height / momentum_roughness_length[~pos]))**2
    )

    # eq. (25)
    stability_factor = np.empty(pos.shape)
    stability_factor[pos] = 1 / (1 + 3 * stability_adjustment_parameter * richardson[pos] * np.sqrt(
        1 + stability_adjustment_parameter * richardson[pos]))
    stability_factor[~pos] = 1 - 3 * stability_adjustment_parameter * richardson[~pos] / (
        1 + c * np.sqrt(-richardson[~pos]))

    stability_factor[np.isnan(stability_factor)] = 1.

    return stability_factor


def calc_turbulent_exchange_coefficient(model):
    """
    Calculate the transfer coefficient C_H from [1] (eq. (22)) including the
    possible adjustment for atmospheric stability.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    s = model.state
    roi = model.grid.roi

    temp_measurement_height = model.config.meteo.measurement_height.temperature
    wind_measurement_height = model.config.meteo.measurement_height.wind

    if model.config.snow.measurement_height_adjustment:
        temp_measurement_height -= s.snow.depth[roi]
        temp_measurement_height = np.maximum(temp_measurement_height, 1.)  # as implemented in FSM

    heat_moisture_roughness_length = 0.1 * s.surface.roughness_length[roi]
    coeff = constants.VON_KARMAN**2 / (
        np.log(wind_measurement_height / s.surface.roughness_length[roi])
        * np.log(temp_measurement_height / heat_moisture_roughness_length)
    )

    if model.config.meteo.stability_correction:
        sf = stability_factor(
            s.meteo.temp[roi],
            s.surface.temp[roi],
            s.meteo.wind_speed[roi],
            s.snow.area_fraction[roi],
            model.config.snow.roughness_length,
            model.config.soil.roughness_length,
            temp_measurement_height,
            wind_measurement_height,
            model.config.meteo.stability_adjustment_parameter,
        )
        coeff *= sf

    s.surface.turbulent_exchange_coeff[roi] = coeff


def calc_radiation_balance(model, pos):
    """
    Calculate outgoing shortwave and longwave radiation and net radiation.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels for which the radiation fluxes should be calculated.
    """
    s = model.state

    # TODO this should be a parameter
    snow_emissivity = 0.99

    s.meteo.sw_out[pos] = s.surface.albedo[pos] * s.meteo.sw_in[pos]
    s.meteo.lw_out[pos] = snow_emissivity * constants.STEFAN_BOLTZMANN * s.surface.temp[pos]**4
    s.meteo.net_radiation[pos] = (
        s.meteo.sw_in[pos]
        - s.meteo.sw_out[pos]
        + s.meteo.lw_in[pos]
        - s.meteo.lw_out[pos]
    )


def calc_saturation_specific_humidity(model, pos):
    """
    Calculate saturation specific humidity at surface temperature.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.
    """
    s = model.state
    sat_vap_press_surf = meteo.saturation_vapor_pressure(s.surface.temp[pos])
    s.surface.sat_spec_hum[pos] = meteo.specific_humidity(
        s.meteo.atmos_press[pos],
        sat_vap_press_surf,
    )


def calc_moisture_availability(model, pos):
    """
    Calculate moisture availability following [1-2].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015

    .. [2] Essery, R. L. H., Best, M. J., & Cox, P. M. (2001). MOSES 2.2
       Technical Documentation, Tech. rep., Hadley Centre, Met Office.
       http://jules.jchmr.org/sites/default/files/HCTN_30.pdf
    """
    s = model.state
    moisture_availability = s.surface.conductance[pos] / (  # eq. (38) from [2]
        s.surface.conductance[pos]
        + s.surface.turbulent_exchange_coeff[pos] * s.meteo.wind_speed[pos]
    )
    moisture_availability[s.surface.sat_spec_hum[pos] < s.meteo.spec_hum[pos]] = 1.
    moisture_availability[s.snow.swe[pos] > 0] = 1.
    s.surface.moisture_availability[pos] = moisture_availability


def calc_latent_heat(model, pos):
    """
    Calculate latent heat.

    Latent heat is set to the latent heat of sublimation for snow covered
    surfaces, and to the latent heat of vaporization elsewhere.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.
    """
    s = model.state
    lat_heat_vap = meteo.latent_heat_of_vaporization(s.meteo.temp[pos])
    s.surface.lat_heat[pos] = np.where(
        (s.snow.ice_content[0, pos] > 0) | (s.surface.temp[pos] < constants.T0),
        lat_heat_vap + constants.LATENT_HEAT_OF_FUSION,
        lat_heat_vap,
    )


def calc_surface_flux(model, pos):
    """
    Calculate the surface flux following [1].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    s = model.state
    s.surface.heat_flux[pos] = (
        2 * s.surface.therm_cond[pos]
        * (s.surface.temp[pos] - s.surface.layer_temp[pos])
        / s.surface.thickness[pos]
    )


def calc_turbulent_fluxes(model, pos, sensible=True, latent=True):
    """
    Calculate turbulent fluxes following [1] (opposed to [1] however, the fluxes
    are here oriented towards the surface).

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.

    sensible : bool, default True
        Calculate the sensible heat flux.

    latent : bool, default True
        Calculate the latent heat flux.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    s = model.state

    rhoa_CH_Ua = (  # (kg m-2 s-1)
        s.meteo.dry_air_density[pos]
        * s.surface.turbulent_exchange_coeff[pos]
        * s.meteo.wind_speed[pos]
    )

    if sensible:
        s.surface.sens_heat_flux[pos] = (
            constants.SPEC_HEAT_CAP_DRY_AIR
            * rhoa_CH_Ua
            * (s.meteo.temp[pos] - s.surface.temp[pos])
        )

    if latent:
        s.surface.moisture_flux[pos] = (
            s.surface.moisture_availability[pos]
            * rhoa_CH_Ua
            * (s.meteo.spec_hum[pos] - s.surface.sat_spec_hum[pos])
        )
        s.surface.lat_heat_flux[pos] = s.surface.lat_heat[pos] * s.surface.moisture_flux[pos]


def calc_advective_heat(model, pos):
    """
    Calculate heat advected by precipitation following [1].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.

    References
    ----------
    .. [1] Strasser, U. (2008). Die Modellierung der Gebirgsschneedecke im
       Nationalpark Berchtesgaden. Modelling of the mountain snow cover in the
       Berchtesgaden National Park, Berchtesgaden National Park research report,
       No. 55, Berchtesgaden.
    """
    s = model.state
    s.surface.advective_heat_flux[pos] = (
        (  # rainfall on snow
            constants.SPEC_HEAT_CAP_WATER
            * (s.meteo.temp[pos] - s.surface.temp[pos])
            * s.meteo.rainfall[pos]
        ) + (  # snowfall on snow
            constants.SPEC_HEAT_CAP_ICE
            * (s.meteo.wet_bulb_temp[pos] - s.surface.temp[pos])
            * s.meteo.snowfall[pos]
        )
    ) / model.timestep


def solve_energy_balance(model, pos):
    """
    Calculate surface temperature and flux changes following [1] (eqs.
    (32)-(35)).

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.

    Returns
    -------
    length-4 ndarray tuple
        Tuple containing the surface temperature change, moisture flux change,
        surface heat flux change and sensible heat flux change.

    References
    ----------
    .. [1] Essery, R. (2015). A factorial snowpack model (FSM 1.0).
       Geoscientific Model Development, 8(12), 3867–3876.
       https://doi.org/10.5194/gmd-8-3867-2015
    """
    s = model.state

    rhoa_CH_Ua = (  # (kg m-2 s-1)
        s.meteo.dry_air_density[pos]
        * s.surface.turbulent_exchange_coeff[pos]
        * s.meteo.wind_speed[pos]
    )

    dQsat_by_dTs = s.surface.lat_heat[pos] * s.surface.sat_spec_hum[pos] / (  # eq. (37) (K-1)
        constants.SPEC_GAS_CONSTANT_WATER_VAPOR * s.surface.temp[pos]**2
    )

    surf_temp_change = (  # eq. (38) plus heat advected by precipitation
        (
            s.meteo.net_radiation[pos]
            + s.surface.sens_heat_flux[pos]
            + s.surface.lat_heat_flux[pos]
            + s.surface.advective_heat_flux[pos]
            - s.surface.heat_flux[pos]
            - constants.LATENT_HEAT_OF_FUSION * (s.snow.melt[pos] / model.timestep)
        ) / (
            (
                constants.SPEC_HEAT_CAP_DRY_AIR
                + s.surface.lat_heat[pos]
                * s.surface.moisture_availability[pos]
                * dQsat_by_dTs
            ) * rhoa_CH_Ua
            + 2 * s.surface.therm_cond[pos] / s.surface.thickness[pos]
            + 4 * constants.STEFAN_BOLTZMANN * s.surface.temp[pos]**3
        )
    )

    surf_moisture_flux_change = (  # eq. (33) (kg m-2 s-1)
        s.surface.moisture_availability[pos]
        * rhoa_CH_Ua
        * dQsat_by_dTs
        * surf_temp_change
    )

    surf_heat_flux_change = (  # eq. (34) (W m-2)
        2 * s.surface.therm_cond[pos]
        * surf_temp_change / s.surface.thickness[pos]
    )

    sens_heat_flux_change = (  # eq. (35) (W m-2)
        constants.SPEC_HEAT_CAP_DRY_AIR
        * rhoa_CH_Ua
        * surf_temp_change
    )

    surf_temp_change[np.isnan(surf_temp_change)] = 0.
    surf_moisture_flux_change[np.isnan(surf_moisture_flux_change)] = 0.
    surf_heat_flux_change[np.isnan(surf_heat_flux_change)] = 0.
    sens_heat_flux_change[np.isnan(sens_heat_flux_change)] = 0.

    return (
        surf_temp_change,
        surf_moisture_flux_change,
        surf_heat_flux_change,
        sens_heat_flux_change,
    )


def energy_balance_remainder(model, pos, surf_temp):
    """
    Update the radiation balance and turbulent fluxes and calculate the
    remainder of the surface energy balance assuming no melt for a given surface
    temperature following [1-2].

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    pos : ndarray(bool)
        Pixels to be considered.

    surf_temp : ndarray
        Surface temperature (K).

    Returns
    -------
    en_bal : ndarray
        Surface energy balance remainder.

    References
    ----------
    .. [1] Strasser, U. (2008). Die Modellierung der Gebirgsschneedecke im
       Nationalpark Berchtesgaden. Modelling of the mountain snow cover in the
       Berchtesgaden National Park, Berchtesgaden National Park research report,
       No. 55, Berchtesgaden.

    .. [2] Hanzer, F., Helfricht, K., Marke, T., & Strasser, U. (2016).
       Multilevel spatiotemporal validation of snow/ice mass balance and runoff
       modeling in glacierized catchments. The Cryosphere, 10(4), 1859–1881.
       https://doi.org/10.5194/tc-10-1859-2016
    """
    s = model.state
    s.surface.temp[pos] = surf_temp

    calc_radiation_balance(model, pos)
    calc_saturation_specific_humidity(model, pos)
    calc_turbulent_fluxes(model, pos)
    calc_advective_heat(model, pos)

    en_bal = (
        s.meteo.net_radiation[pos]
        + s.surface.sens_heat_flux[pos]
        + s.surface.lat_heat_flux[pos]
        + s.surface.advective_heat_flux[pos]
        - s.surface.heat_flux[pos]
    )

    return en_bal


def iterate_surface_temperature(model, frosties):
    """
    Iterate the snow surface temperature to find the zeros of the energy balance
    equation assuming no melt.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    frosties : ndarray(bool)
        Pixels to be considered.
    """
    if not frosties.any():
        return

    s = model.state
    method = model.config.snow.cryolayers.surface_temperature_iteration_method

    if method == 'legacy':
        roi = model.grid.roi
        en_bal = np.zeros(roi.shape)
        iteraties = frosties
        max_temp = constants.T0
        min_temp = s.meteo.temp[frosties].min() - 3.
        temp_inc = -0.25
        for surf_temp_iter in np.arange(max_temp, min_temp - 1e-6, temp_inc):
            en_bal[iteraties] = energy_balance_remainder(model, iteraties, surf_temp_iter)
            iteraties = model.roi_mask_to_global(frosties[roi] & (en_bal[roi] < 0.))
    elif method == 'secant':
        tol = 1e-2

        iteraties = frosties.copy()
        iteraties_idxs = np.where(iteraties.flat)[0]
        x0 = np.full(len(iteraties_idxs), constants.T0 - 10.)
        x1 = np.full(len(iteraties_idxs), constants.T0)
        y0 = energy_balance_remainder(model, iteraties, x0)
        y1 = energy_balance_remainder(model, iteraties, x1)

        while True:
            d = (x1 - x0) / (y1 - y0) * y1  # secant method
            iter_pos = np.abs(d) > tol

            if iter_pos.sum() == 0:
                break

            d = d[iter_pos]
            x0 = x0[iter_pos]
            x1 = x1[iter_pos]
            y0 = y0[iter_pos]
            y1 = y1[iter_pos]

            x0 = x1.copy()
            y0 = y1.copy()
            x1 -= d
            iteraties.flat[iteraties_idxs[~iter_pos]] = False
            y1 = energy_balance_remainder(model, iteraties, x1)
            iteraties_idxs = iteraties_idxs[iter_pos]

        # TODO calculate melt when surface temperature is positive after iteration?
        s.surface.temp[frosties] = s.surface.temp[frosties].clip(max=constants.T0)



import numpy as np
from openamundsen import constants, errors, util
import pandas as pd
import rasterio
import xarray as xr


_POINT_DATASET_META_VARS = [
    'station_name',
    'lon',
    'lat',
    'alt',
]
_POINT_DATASET_ALLOWED_METEO_VARS = list(constants.METEO_VAR_METADATA.keys())
_POINT_DATASET_GRID_VARS = [
    'x',
    'y',
    'col',
    'row',
    'within_grid_extent',
    'within_roi',
]
_POINT_DATASET_MINIMAL_VARS = _POINT_DATASET_META_VARS + constants.MINIMUM_REQUIRED_METEO_VARS


def make_point_dataset(
        data=None,
        dates=None,
        point_id=None,
        name=None,
        lon=None,
        lat=None,
        alt=None,
):
    """
    Create a dataset containing the meteorological forcing data for a single
    point including metadata (point ID and name, variable units).

    Parameters
    ----------
    data : xr.Dataset or pd.DataFrame, optional
        Point data. If None, a dataset containing all-nan values will be
        created.

    dates : pd.DatetimeIndex, optional
        Must be specified only if `data` is None, otherwise the dates will be
        taken from there.

    point_id : str, optional
        Point ID. Only required if the ID is not stored in `data.attrs`.

    name : str, optional
        Point name. If `None`, the name is taken from `data.attrs` if possible,
        otherwise the name will be set to the point ID.

    lon : float, optional
        Longitude (degrees). Must be specified only if `data` is None or does
        not contain a `lon` variable.

    lat : float, optional
        Latitude (degrees). Must be specified only if `data` is None or does
        not contain a `lat` variable.

    alt : float, optional
        Altitude (m). Must be specified only if `data` is None or does
        not contain an `alt` variable.

    Returns
    -------
    ds : xr.Dataset
    """
    if data is None:
        data = xr.Dataset()

        if dates is None:
            dates = pd.DatetimeIndex([])
    elif isinstance(data, xr.Dataset):
        data = data.copy(deep=True)  # required because of the potential precipitation-altering
        dates = data.indexes['time']
    elif isinstance(data, pd.DataFrame):
        data = data.copy()
        data.index.name = 'time'
        dates = data.index
        data = data.to_xarray()
    else:
        raise TypeError(f'Unsupported type: {type(data)}')

    if point_id is None:
        if data is None or 'station_id' not in data.attrs:
            raise TypeError('Point ID neither in data nor passed as keyword argument')
        point_id = data.attrs['station_id']

    if name is None:
        try:
            name = data.attrs['station_name']
        except KeyError:
            name = point_id

    if lon is None:
        if 'lon' not in data.variables:
            raise TypeError('Longitude neither in data nor passed as keyword argument')
        lon = float(data.lon)

    if lat is None:
        if 'lat' not in data.variables:
            raise TypeError('Latitude neither in data nor passed as keyword argument')
        lat = float(data.lat)

    if alt is None:
        if 'alt' not in data.variables:
            raise TypeError('Altitude neither in data nor passed as keyword argument')
        alt = float(data.alt)

    meteo_vars = list(set(data.data_vars) & set(_POINT_DATASET_ALLOWED_METEO_VARS))
    ds = make_empty_point_dataset(
        dates,
        point_id,
        name,
        lon,
        lat,
        alt,
        meteo_vars=meteo_vars,
    )

    # Copy values (and check if the units are as expected)
    for var in meteo_vars:
        meta = constants.METEO_VAR_METADATA[var]

        if var == 'precip':
            # Convert precipitation rates to sums first if required
            if 'units' in data['precip'].attrs and data['precip'].units == 'kg m-2 s-1':
                try:
                    freq = pd.infer_freq(dates)
                    dt = util.offset_to_timedelta(freq).total_seconds()
                except ValueError:  # infer_freq() needs at least 3 dates
                    dt = np.nan

                data['precip'] *= dt
                data['precip'].attrs['units'] = 'kg m-2'
        elif var == 'cloud_fraction':
            # Convert cloud cover fraction to percent if required: either if the units are
            # explicitly set to "1" or if no units are set but the maximum cloud fraction value is
            # <= 1. (As the cloud_fraction model variable has units "1" but for the forcing data
            # units "%" are expected, this could happen easily.)
            if (
                data['cloud_fraction'].attrs.get('units') == '1'
                or (len(data) > 10 and data['cloud_fraction'].max() <= 1)
            ):
                data['cloud_fraction'] *= 100
                data['cloud_fraction'].attrs['units'] = '%'

        units = meta['units']
        if 'units' in data[var].attrs and data[var].units != units:
            raise errors.MeteoDataError(
                f'{var} has wrong units: expected {units}, got {ds[var].units}'
            )

        ds[var].values[:] = data[var].values

    return ds


def make_empty_point_dataset(dates, point_id, name, lon, lat, alt, meteo_vars=None):
    """
    Create an empty dataset for the meteorological forcing data for a single
    point.

    The returned dataset contains the station metadata (point ID, name,
    coordinates and altitude) and the meteorological variables, which are set to
    all-nan values.

    Parameters
    ----------
    dates : pd.DatetimeIndex
        Dates.

    point_id : str
        Point ID.

    name : str
        Point name

    lon : float
        Longitude (degrees).

    lat : float
        Latitude (degrees).

    alt : float
        Altitude (m)

    meteo_vars : list, default None
        Meteorological variables to create. If None, create a dataset with the
        minimally required variables.

    Returns
    -------
    ds : xr.Dataset
    """
    data = {
        'lon': ([], float(lon)),
        'lat': ([], float(lat)),
        'alt': ([], float(alt)),
    }

    if meteo_vars is None:
        meteo_vars = constants.MINIMUM_REQUIRED_METEO_VARS

    for var in meteo_vars:
        meta = constants.METEO_VAR_METADATA[var]
        data[var] = (
            ['time'],
            np.full(len(dates), np.nan),
            meta,
        )

    ds = xr.Dataset(
        data,
        coords={'time': (['time'], dates)},
        attrs={
            'station_id': point_id,
            'station_name': name,
        },
    )

    return ds


def combine_point_datasets(datasets, add_minimum_required_vars=True):
    """
    Combine a list of meteo datasets as returned by `make_point_dataset`.
    The datasets are merged by adding an additional "station" (= station id)
    dimension.

    Parameters
    ----------
    add_minimum_required_vars : bool, default True
        Add the minimum required variables for an openAMUNDSEN model run to the
        resulting combined dataset even if the variables do not exist in any of
        the individual datasets.
    """
    datasets_processed = []

    # Create a list of all meteo variables occurring in at least one of the datasets
    meteo_vars = set()
    if add_minimum_required_vars:
        meteo_vars.update(set(constants.MINIMUM_REQUIRED_METEO_VARS))
    for ds in datasets:
        meteo_vars.update(set(ds.data_vars))
    meteo_vars = sorted(list(meteo_vars & set(_POINT_DATASET_ALLOWED_METEO_VARS)))

    for ds in datasets:
        ds = ds.copy()

        for var in meteo_vars:
            meta = constants.METEO_VAR_METADATA[var]

            # Add missing variables to the dataset
            if var not in ds:
                ds[var] = xr.DataArray(
                    data=np.full(ds.time.shape, np.nan),
                    coords=ds.coords,
                    dims=('time',),
                )

            ds[var].attrs = {
                'standard_name': meta['standard_name'],
                'units': meta['units'],
            }

        ds.coords['station'] = ds.attrs['station_id']
        ds['station_name'] = ds.attrs['station_name']
        ds.attrs = {}

        datasets_processed.append(ds)

    return xr.combine_nested(datasets_processed, concat_dim='station')


def prepare_point_coordinates(ds, grid, crs):
    """
    Transform the lon/lat coordinates of the meteorological stations to the
    coordinate system of the model grid. The transformed coordinates are
    stored in the `x` and `y` variables of the meteo dataset.
    Additionally, the row and column indices of the stations within the
    model grid are stored in the `row` and `col` variables, and two boolean
    variables `within_grid_extent` and `within_roi` indicate whether the
    stations lie within the model grid extent and the ROI, respectively.

    Parameters
    ----------
    ds : xr.Dataset
        Point forcing dataset as returned by `combine_point_datasets`.

    grid : ModelGrid
        Model grid.

    crs : str
        Coordinate reference system.

    Returns
    -------
    ds : xr.Dataset
    """
    ds = ds.copy()

    x, y = util.transform_coords(ds.lon, ds.lat, constants.CRS_WGS84, crs)

    x_var = ds.lon.copy()
    x_var.values = x
    x_var.attrs = {
        'standard_name': 'projection_x_coordinate',
        'units': 'm',
    }

    y_var = ds.lat.copy()
    y_var.values = y
    y_var.attrs = {
        'standard_name': 'projection_y_coordinate',
        'units': 'm',
    }

    ds['x'] = x_var
    ds['y'] = y_var

    bool_var = ds.x.copy().astype(bool)
    bool_var[:] = False
    bool_var.attrs = {}
    int_var = bool_var.copy().astype(int)
    int_var[:] = -1

    if len(x) > 0:
        rows, cols = rasterio.transform.rowcol(grid.transform, x, y)
        rows = np.array(rows).astype(int)
        cols = np.array(cols).astype(int)
    else:
        rows = []
        cols = []

    row_var = int_var.copy()
    col_var = int_var.copy()
    row_var.values[:] = rows
    col_var.values[:] = cols
    ds['col'] = col_var
    ds['row'] = row_var

    ds['within_grid_extent'] = (
        (col_var >= 0)
        & (col_var < grid.cols)
        & (row_var >= 0)
        & (row_var < grid.rows)
    )

    within_roi_var = bool_var.copy()
    ds['within_roi'] = within_roi_var

    for station in ds.indexes['station']:
        dss = ds.sel(station=station)

        if dss.within_grid_extent:
            row = int(dss.row)
            col = int(dss.col)
            within_roi_var.loc[station] = grid.roi[row, col]

    # reorder variables (only for aesthetic reasons)
    meteo_vars = sorted(list(set(_POINT_DATASET_ALLOWED_METEO_VARS) & set(ds.data_vars)))
    var_order = _POINT_DATASET_META_VARS + _POINT_DATASET_GRID_VARS + meteo_vars
    ds = ds[var_order]

    return ds


def strip_point_dataset(ds):
    """
    Remove the grid-specific variables from a point forcing dataset.
    """
    return ds.drop_vars(_POINT_DATASET_GRID_VARS)


def is_valid_point_dataset(ds, dates=None):
    """
    Test if the passed variable is a valid point forcing dataset.
    """
    if set(ds.dims) != set(['station', 'time']):
        return False

    for v in _POINT_DATASET_MINIMAL_VARS:
        if v not in ds:
            return False

    if dates is not None:
        if not pd.DatetimeIndex(dates).equals(ds.indexes['time']):
            return False

    return True
